"""DSAUploader component and UploadType 
"""
import os
import time
import shutil

import json
import numpy as np
import pandas as pd
import requests
from flask import request

from typing_extensions import Union

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, Patch, no_update, dash_table
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
#import dash_treeview_antd as dta
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform
import dash_uploader as du

from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools.utils.shapes import load_annotations

from fusion_tools import DSATool
from fusion_tools.handler.plugin import DSAPluginRunner, DSAPluginGroup
from fusion_tools.handler.resource_selector import DSAResourceSelector

from girder_job_sequence import Job, Sequence
from girder_job_sequence.utils import from_list, from_dict
import threading
import large_image


# Maximum allowed size of uploads (Mb)
MAX_UPLOAD_SIZE = 1e4
# Minimum chunk size (Mb)
MIN_UPLOAD_SIZE = 6

WSI_TYPES = [i for i in list(large_image.listSources()['extensions'].keys()) if not i in ['json','yaml','yml']]
ANN_TYPES = ['json','geojson','xml','csv']


class DSAUploadType:
    """Formatted upload type for a DSAUploader Component.
    """
    def __init__(self,
                 name: str,
                 description: str,
                 input_files: list = [],
                 processing_plugins:Union[list,None] = None,
                 required_metadata: Union[list,None] = None):
        """Constructor method

        :param name: Name for this upload type (appears in dropdown menu in DSAUploader component)
        :type name: str
        :param input_files: List of dictionaries containing the following keys: name:str, description: str, accepted_types: Union[list,None], preprocessing_plugins: Union[list,None], main: bool, required: bool
        :type input_files: list
        :param processing_plugins: List of plugins to run after data has been uploaded. Allows for input of plugin-specific arguments after completion of file uploads., defaults to None
        :type processing_plugins: Union[list,None], optional
        :param required_metadata: List of "keys" which require user input either by uploading a file or by manual addition. Can either be a list of strings or a list of dicts with keys 'key':str,'values':list,'required':bool and strings (assumed required=True)
        :type required_metadata: Union[list,None], optional
        """
        
        self.name = name
        self.description = description

        # At least one element in input_files must contain 'main': True and 'required': True 
        # (just to ensure there has to be something uploaded and it has to at least be the main file)
        self.input_files = input_files
        self.processing_plugins = processing_plugins
        self.required_metadata = required_metadata

        # At a minimum, input files just has to contain at least one element and at least one "required" element
        # Other acceptable keys include "type", "description", "accepted_types", and "parent"
        assert len(self.input_files)>0
        assert all([type(i) in [dict,str] for i in self.input_files])

        default_vals = {
            'accepted_types': None,
            'preprocessing_plugins': None,
            'type': 'item',
            'parent': None,
            'required': True
        }

        # Checking the input file types
        files = [i for i in self.input_files if type(i)==dict]        
        assert all(['name' in i for i in files])
        files = [i for i in self.input_files if 'type' in files]
        assert all([i['type'] in ['item','file','annotation'] for i in files])

        # Checking that all the file and annotation types have a "parent"
        files = [i for i in files if i['type'] in ['file','annotation']]
        assert all(['parent' in i for i in files])

        for idx,i in enumerate(self.input_files):
            if type(i)==dict:
                if not all([j in i for j in default_vals]):
                    for key,val in default_vals.items():
                        if not key in i:
                            i[key] = val
            elif type(i)==str:
                i_dict = {
                    'name': i
                } | default_vals
                self.input_files[idx] = i_dict
        
        
        # Checking format of required_metadata (if it's a dictionary it has to have 'name'and 'required')
        if not self.required_metadata is None:
            req_meta_check = []
            for r in self.required_metadata:
                if type(r)==str:
                    req_meta_check.append(True)
                elif type(r)==dict:
                    req_meta_check.append(
                        all([i in r for i in ['name', 'required', 'item']])
                    )
        else:
            req_meta_check = [True]
            self.required_metadata = []

        assert all(req_meta_check)

        # All processing_plugins have to have a "dict" or "list" type
        assert all([isinstance(i,dict) or isinstance(i,list) for i in self.processing_plugins])

        # Lists within self.processing_plugins are interpreted as a sequence
        # Jobs that aren't in a sequence are determined to be independent of other processing steps
        # and are therefore executed at the same time as the first job in each sequence.
        """
        self.job_sequences = []
        for p in self.processing_plugins:
            if type(p)==dict:
                self.job_sequences.append(from_dict(p))
            elif type(p)==list:
                self.job_sequences.append(from_list(p))
        """

# RequestData is from the documentation for dash-uploader with the keys modified from "flow" to "resumable" (there must have been some update that they didn't account for initially)
class RequestData:
    # A helper class that contains data from the request
    # parsed into handier form.

    def __init__(self, request):
        """
        Parameters
        ----------
        request: flask.request
            The Flask request object
        """
        # Available fields: https://github.com/flowjs/flow.js
        self.n_chunks_total = request.form.get("resumableTotalChunks", type=int)
        self.chunk_number = request.form.get("resumableChunkNumber", default=1, type=int)
        self.filename = request.form.get("resumableFilename", default="error", type=str)
        self.total_size = request.form.get('resumableTotalSize',default=0,type=int)
        # 'unique' identifier for the file that is being uploaded.
        # Made of the file size and file name (with relative path, if available)
        self.unique_identifier = request.form.get(
            "resumableIdentifier", default="error", type=str
        )
        # flowRelativePath is the flowFilename with the directory structure included
        # the path is relative to the chosen folder.
        self.relative_path = request.form.get("resumableRelativePath", default="", type=str)
        if not self.relative_path:
            self.relative_path = self.filename

        # Get the chunk data.
        # Type of `chunk_data`: werkzeug.datastructures.FileStorage
        self.chunk_data = request.files["file"]
        self.upload_id = request.form.get("upload_id", default="", type=str)

class DSAUploadHandler:
    def __init__(self,server,upload_folder,use_upload_id):
        self.server = server
        self.upload_folder = upload_folder
        self.use_upload_id = use_upload_id

        # 2**20 is 1MB in binary
        self.min_chunk_size = MIN_UPLOAD_SIZE * (2**20)

    def post_before(self,req_data):
        # Creating item to upload to
        upload_info = json.loads(req_data.upload_id)

        if upload_info['fusion_upload_type'] in ['item','file']:
            print(f'Creating file: {req_data.filename}')
            post_response = requests.post(
                upload_info['api_url']+'/file',
                params = {
                    'token': upload_info['token'],
                    'parentType': upload_info['parentType'],
                    'parentId': upload_info['parentId'],
                    'name': req_data.filename,
                    'size': req_data.total_size
                }
            )

            self.current_upload = post_response.json()
        else:
            self.current_upload = []

    def post_file_chunk(self,api_url,token,parentId,chunk,offset):

        post_response = requests.post(
            api_url+f'/file/chunk',
            params={
                'token': token,
                'uploadId': parentId,
                'offset': offset * self.min_chunk_size
            },
            data = chunk
        )
        return post_response.json()

    def remove_file(self,path):
        os.unlink(path)
    
    def save_annotation_chunk(self):

        r = RequestData(request)

        # Getting temporary directory containing chunks
        tmp_chunk_path = self.upload_folder / r.unique_identifier
        if not tmp_chunk_path.exists():
            tmp_chunk_path.mkdir(parents=True)

        chunk_name = f'{r.filename}_{r.chunk_number}'
        chunk_file = tmp_chunk_path / chunk_name

        # Making a lock file
        lock_file_path = tmp_chunk_path / f'.lock_{r.chunk_number}'

        with open(lock_file_path,'a'):
            os.utime(lock_file_path,None)

        # Save the chunk, delete the lock
        r.chunk_data.save(chunk_file)
        self.remove_file(lock_file_path)

        # Check if all chunks are present
        upload_complete = all([os.path.exists(os.path.join(tmp_chunk_path,f'{r.filename}_{i}')) for i in range(1,r.n_chunks_total+1)])

        # Reassembling if the upload is complete
        if upload_complete:
            # Wait until all the lock files are removed
            wait = 0
            while any([os.path.isfile(os.path.join(tmp_chunk_path,f'.lock_{i}')) for i in range(1,r.n_chunks_total+1)]):
                wait +=1
                if wait>=5:
                    raise Exception(
                        f"Error uploading files to {tmp_chunk_path}"
                    )
                time.sleep(1)
            
            target_file_path = os.path.join(self.upload_folder,r.filename)
            if os.path.exists(target_file_path):
                # dash-uploader implements a retry() callback for this if it doesn't go through the first time
                self.remove_file(target_file_path)
            
            with open(target_file_path,'ab') as target_file:
                chunk_paths = [os.path.join(tmp_chunk_path,f'{r.filename}_{i}') for i in range(1,r.n_chunks_total+1)]
                for p in chunk_paths:
                    with open(p,'rb') as stored_chunk:
                        target_file.write(stored_chunk.read())
            
            shutil.rmtree(tmp_chunk_path)

    def post_annotations(self, upload_info):

        r = RequestData(request)
        processed_annotations = load_annotations(os.path.join(self.upload_folder,r.filename))
        self.remove_file(os.path.join(self.upload_folder,r.filename))

        response = requests.post(
            upload_info['api_url']+f'/annotation/item/{upload_info["parentId"]}?token={upload_info["token"]}',
            data = json.dumps(processed_annotations),
            headers = {
                'X-HTTP-Method': 'POST',
                'Content-Type': 'application/json'
            }
        )

        return response.json()

    def post(self):

        r = RequestData(request)
        upload_info = json.loads(r.upload_id)

        if r.chunk_number==1:
            self.post_before(r)
        
        if upload_info['fusion_upload_type'] in ['item','file']:
            post_response = self.post_file_chunk(
                api_url=upload_info['api_url'],
                token = upload_info['token'],
                parentId=self.current_upload['_id'],
                chunk = r.chunk_data.read(),
                offset = r.chunk_number-1
            )
        else:

            self.save_annotation_chunk()

            if r.chunk_number==r.n_chunks_total:
                post_response = self.post_annotations(
                    upload_info
                )
                post_response = {
                    'n_annotations': post_response,
                    '_modelType': 'annotation'
                }
            else:
                post_response = r.filename

        return json.dumps(post_response | upload_info)

    def get_before(self):
        pass

    def get(self):
        return 200

    def get_after(self):
        pass

    def post_after(self):
        pass


class DSAUploader(DSATool):
    """Handler for DSAUploader component, handling uploading data to a specific folder, adding metadata, and running sets of preprocessing plugins.

    :param DSATool: Sub-class of Tool specific to DSA components. Updates with session data by default.
    :type DSATool: None
    """
    def __init__(self,
                 handler,
                 dsa_upload_types: Union[DSAUploadType,list] = []):
        

        super().__init__()
        self.handler = handler
        self.dsa_upload_types = dsa_upload_types

    def __str__(self):
        return 'DSA Uploader'

    def load(self,component_prefix:int):

        self.component_prefix = component_prefix

        self.title = 'Dataset Uploader'
        self.blueprint = DashBlueprint(
            transforms=[
                PrefixIdTransform(prefix=f'{component_prefix}',escape = lambda input_id: self.prefix_escape(input_id)),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

        # Loading resource selector
        #self.resource_selector = DSAResourceSelector(
        #    handler = self.handler
        #)

        #self.resource_selector.load(self.component_prefix)

        self.plugin_inputs_handler = DSAPluginGroup(
            handler = self.handler
        )
        self.plugin_inputs_handler.load(self.component_prefix)
        self.plugin_inputs_handler.gen_layout({})

    def update_layout(self, session_data:dict, use_prefix:bool):

        if not 'current_user' in session_data:
            uploader_children = html.Div(
                dbc.Alert(
                    'Make sure to login first in order to upload!',
                    color = 'warning'
                )
            )
   
        else:
            self.plugin_inputs_handler.update_layout(session_data,use_prefix=use_prefix)
            uploader_children = html.Div([
                dbc.Row([
                    #dbc.Modal(
                    #    id = {'type': 'dsa-uploader-selector-modal','index': 0},
                    #    centered = True,
                    #    is_open = False,
                    #    size = 'xl',
                    #    className = None,
                    #    children = [
                    #        self.resource_selector.blueprint.embed(self.blueprint)
                    #    ]
                    #),
                    dcc.Loading(
                        html.Div(
                            id = {'type': 'dsa-uploader-collection-or-user-div','index': 0},
                            children = [
                                html.Div(
                                    id = {'type': 'dsa-uploader-folder-nav-parent','index': 0},
                                    children = []
                                ),
                                dbc.Stack([
                                    dbc.Button(
                                        'Collection',
                                        n_clicks = 0,
                                        className = 'd-grid col-6 mx-auto',
                                        color = 'primary',
                                        id = {'type': 'dsa-uploader-collection-button','index': 0}
                                    ),
                                    dbc.Button(
                                        'User Folder',
                                        n_clicks = 0,
                                        className = 'd-grid col-6 mx-auto',
                                        color = 'secondary',
                                        id = {'type': 'dsa-uploader-user-folder-button','index': 0}
                                    )
                                ],direction = 'horizontal',gap = 3)
                            ]
                        )
                    ),
                    html.Div(
                        id = {'type': 'dsa-uploader-new-folder-div','index': 0},
                        children = []
                    ),
                    html.Div(
                        id = {'type': 'dsa-uploader-new-folder-error-div','index': 0},
                        children = []
                    )
                ]),
                html.Hr(),
                dbc.Row([
                    html.Div(
                        id = {'type': 'dsa-uploader-upload-type-div','index': 0},
                        children = []
                    )
                ]),
                html.Hr(),
                dbc.Row([
                    html.Div(
                        id = {'type': 'dsa-uploader-processing-plugins-div','index': 0},
                        children = []
                    )
                ])
            ])

        layout = html.Div([
            dbc.Card(
                dbc.CardBody([
                    dbc.Row(
                        html.H3('Dataset Uploader')
                    ),
                    html.Hr(),
                    dbc.Row(
                        'Uploading slides and associated files to a particular folder on attached DSA instance. Access pre-processing plugins.'
                    ),
                    html.Hr(),
                    html.Div(
                        self.plugin_inputs_handler.blueprint.embed(self.blueprint),
                        style = {'display':'none'}
                    ),
                    uploader_children
                ])
            )
        ],style = {'maxHeight': '90vh','overflow': 'scroll'})

        if use_prefix:
            PrefixIdTransform(prefix=self.component_prefix,escape = lambda input_id: self.prefix_escape(input_id)).transform_layout(layout)

        return layout

    def gen_layout(self,session_data:Union[dict,None]):

        self.blueprint.layout = self.update_layout(session_data,use_prefix=False)
        
    def get_callbacks(self):

        # Callback for running processing plugin with inputs

        # Callback for selecting whether to upload to public/private collection or user public/private folder
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-collection-button','index': ALL},'n_clicks'),
                Input({'type': 'dsa-uploader-user-folder-button','index': ALL},'n_clicks'),
                Input({'type': 'dsa-uploader-folder-table','index': ALL},'selected_rows'),
                Input({'type': 'dsa-uploader-folder-crumb','index': ALL},'n_clicks'),
                Input({'type': 'dsa-uploader-folder-back-icon','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-uploader-folder-table','index': ALL},'data'),
                State({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-uploader-collection-or-user-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children'),
                Output({'type': 'dsa-uploader-new-folder-div','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.populate_folder_div)

        # Callback for creating a new folder at a specific location
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-new-folder-button','index': ALL},'n_clicks'),
                Input({'type': 'dsa-uploader-new-folder-submit-button','index': ALL},'n_clicks'),
                Input({'type': 'dsa-uploader-new-folder-cancel-button','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-uploader-new-folder-input','index': ALL},'value'),
                State({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-uploader-new-folder-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-new-folder-error-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-folder-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children'),
                Output({'type': 'dsa-uploader-new-folder-button','index': ALL},'disabled')
            ],
            prevent_initial_call = True
        )(self.populate_new_folder)

        # Callback for after selecting a folder to upload to, populating dropdown menu with provided upload types
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-select-folder','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children')
            ],
            [
                Output({'type': 'dsa-uploader-upload-type-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children'),
                Output({'type': 'dsa-uploader-new-folder-button','index': ALL},'disabled'),
                Output({'type': 'dsa-uploader-select-folder','index': ALL},'disabled'),
                Output({'type': 'dsa-uploader-folder-back-icon-div','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.populate_upload_type)

        # Callback for populating with DSAUploadType specifications
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-upload-type-drop','index': ALL},'value')
            ],
            [
                State({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-uploader-upload-type-description-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-upload-type-upload-files-div','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.make_file_uploads)

        # Callback for enabling "Done" button when all required files are uploaded and enabling child uploads of items
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-file-upload','index': ALL},'isCompleted')
            ],
            [
                State({'type': 'dsa-uploader-upload-type-drop','index': ALL},'value'),
                State({'type': 'dsa-uploader-file-upload','index': ALL},'fileNames'),
                State('anchor-vis-store','data'),
                State({'type': 'dsa-uploader-upload-files-store','index': ALL},'data')
            ],
            [
                Output({'type': 'dsa-uploader-file-upload-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-file-upload-done-button','index': ALL},'disabled'),
                Output({'type': 'dsa-uploader-upload-type-drop','index': ALL},'disabled'),
                Output({'type': 'dsa-uploader-upload-files-store','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.enable_upload_done)

        # Callback for populating processing-plugins div after "Done" button is clicked
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-file-upload-done-button','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-uploader-upload-type-drop','index': ALL},'value'),
                State({'type': 'dsa-uploader-folder-nav-parent','index': ALL},'children'),
                State({'type':'dsa-uploader-upload-files-store','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-uploader-processing-plugins-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-upload-type-drop','index': ALL},'disabled'),
                Output({'type': 'dsa-uploader-file-upload-div','index': ALL},'children'),
                Output({'type': 'dsa-uploader-file-upload-done-button','index': ALL},'disabled')
            ],
            prevent_initial_call = True
        )(self.populate_processing_plugins)

        # Callback for checking if required metadata rows are populated
        #TODO: Submitting metadata depends on if the "item" that the metadata is attached to is uploaded yet
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-metadata-table','index': ALL},'data')
            ],
            [
                State({'type': 'dsa-uploader-upload-type-drop','index': ALL},'value')
            ],
            [
                Output({'type': 'dsa-uploader-metadata-submit-button','index': ALL},'disabled')
            ],
            prevent_initial_call = True
        )(self.enable_submit_metadata)

        # Callback for adding new row to custom metadata table
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-metadata-table-add-row','index': MATCH},'n_clicks')
            ],
            [
                State({'type': 'dsa-uploader-metadata-table','index': MATCH},'data')
            ],
            [
                Output({'type': 'dsa-uploader-metadata-table','index': MATCH},'data')
            ],
            prevent_initial_call = True
        )(self.add_row_custom_metadata)

        # Callback for submitting metadata
        self.blueprint.callback(
            [
                Input({'type': 'dsa-uploader-metadata-submit-button','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-uploader-upload-files-store','index': ALL},'data'),
                State({'type': 'dsa-uploader-metadata-table','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-uploader-metadata-submit-status-div','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.submit_metadata)

    def make_selectable_dash_table(self, dataframe:pd.DataFrame, id:dict, multi_row:bool = True, selected_rows: list = []):
        """Generate a selectable DataTable to add to the layout

        :param dataframe: Pandas DataFrame containing columns/rows of interest
        :type dataframe: pd.DataFrame
        :param id: Dictionary containing "type" and "index" keys for interactivity
        :type id: dict
        :param multi_row: Whether to allow selection of multiple rows in the table or just single, defaults to True
        :type multi_row: bool, optional
        :return: dash_table.DataTable component to be added to layout
        :rtype: dash_table.DataTable
        """

        #Optional: Can hide "ID" columns by adding any column containing "ID" to a list of "hidden_columns"

        dataframe = pd.json_normalize(dataframe.to_dict('records'))
        selectable_table = dash_table.DataTable(
            id = id,
            columns = [{'name':i,'id':i,'deletable':False} for i in dataframe.columns],
            data = dataframe.to_dict('records'),
            editable = False,
            filter_action='native',
            sort_action = 'native',
            sort_mode = 'multi',
            column_selectable = 'single',
            row_selectable = 'multi' if multi_row else 'single',
            row_deletable = False,
            selected_rows = selected_rows,
            page_action='native',
            page_current=0,
            page_size=10,
            style_cell = {
                'overflow':'hidden',
                'textOverflow':'ellipsis',
                'maxWidth':0                
            },
            tooltip_data = [
                {
                    column: {'value':str(value),'type':'markdown'}
                    for column, value in row.items()
                } for row in dataframe.to_dict('records')
            ],
            tooltip_duration = None
        )

        return selectable_table

    def organize_folder_contents(self, folder_info:dict, show_empty:bool=True, ignore_histoqc:bool=True)->list:
        """For a given folder selection, return a list of slides(0th) and folders (1th)

        :param folder_info: Folder info dict returned by self.handler.get_path_info(path)
        :type folder_info: dict
        :param show_empty: Whether or not to display folders which contain 0 slides, defaults to False
        :type show_empty: bool, optional
        :return: List of slides within the current folder as well as folders within that folder
        :rtype: list
        """

        folder_folders = []
        folder_slides = []

        # Starting with slides (which will report parent folderId but not that parent's folderId (if applicable))
        if folder_info['_modelType'] in ['folder','collection']:
            all_folder_slides = self.handler.get_folder_slides(
                folder_path = folder_info['_id'],
                folder_type = folder_info['_modelType'],
                ignore_histoqc=ignore_histoqc
            )

            folder_slides_folders = [i['folderId'] for i in all_folder_slides]
            unique_folders = list(set(folder_slides_folders))
            folders_in_folder = []
            for u in unique_folders:
                if not u==folder_info['_id'] and not u in folders_in_folder:
                    # This is for all folders in this folder
                    # This grabs parent folders of this folder
                    u_folder_info = self.handler.get_folder_info(folder_id=u)
                    u_folder_rootpath = self.handler.get_folder_rootpath(u)
                    # Folders in order from collection-->child folder-->etc.
                    folder_ids = [i['object']['_id'] for i in u_folder_rootpath]

                    if any([i['object']['_modelType']=='collection' for i in u_folder_rootpath]):
                        base_model = 'collection'
                    else:
                        base_model = 'user'

                    if folder_ids[-1]==folder_info['_id']:
                        child_folder_path = f'/{base_model}/'+'/'.join([i['object']['name'] if i['object']['_modelType'] in ['folder','collection'] else i['object']['login'] for i in u_folder_rootpath]+[u_folder_info['name']])
                    else:
                        # Folder that is immediate child of current folder:
                        child_folder_idx = folder_ids.index(folder_info['_id'])
                        child_folder_path = f'/{base_model}/'+'/'.join([i['object']['name'] if i['object']['_modelType'] in ['folder','collection'] else i['object']['login'] for i in u_folder_rootpath[:child_folder_idx+2]])


                    child_folder_path_info = self.handler.get_path_info(
                        path = child_folder_path
                    )
                    if not child_folder_path_info['_id'] in folders_in_folder:
                        folders_in_folder.append(child_folder_path_info['_id'])
                        
                        # Adding folder to list if the number of items is above zero or show_empty is True
                        folder_folders.append({
                            'Name': child_folder_path_info['name'],
                            'Folder ID': child_folder_path_info['_id'],
                            'Number of Folders': child_folder_path_info['nFolders'],
                            'Number of Slides': child_folder_path_info['nItems'],
                            'Last Updated': child_folder_path_info['updated']
                        } | child_folder_path_info['meta'])

                elif u==folder_info['_id']:
                    # This means that there are some slides that are direct children (not in a sub-folder) in this folder. 
                    # This adds them all at once
                    for i in all_folder_slides:
                        if i['folderId']==folder_info['_id']:
                            folder_slides.append(
                                {
                                    'Slide Name': i['name'],
                                    'Slide ID': i['_id'],
                                    'Last Updated':i['updated']
                                } | {k:v for k,v in i['meta'].items() if type(v)==str}
                            )

        else:
            
            user_folders = ['Private','Public']
            for u_f in user_folders:
                user_folder_info = self.handler.get_path_info(
                    path = f'/user/{folder_info["login"]}/{u_f}'
                )

                folder_folders.append({
                    'Name': user_folder_info['name'],
                    'Folder ID': user_folder_info['_id'],
                    'Number of Folders': user_folder_info['nFolders'],
                    'Number of Slides': user_folder_info['nItems'],
                    'Last Updated': user_folder_info['updated']
                })

        if show_empty:
            # This is how you get all the empty folders within a folder (does not get child empty folders)
            empty_folders = self.handler.get_folder_folders(
                folder_id = folder_info['_id'],
                folder_type = folder_info['_modelType']
            )
            
            for f in empty_folders:
                if not f['_id'] in folders_in_folder and not f['_id'] in unique_folders:
                    folder_info = self.handler.gc.get(f'/folder/{f["_id"]}/details')
                    folder_folders.append(
                        {
                            'Name': f['name'],
                            'Folder ID': f['_id'],
                            'Number of Folders': folder_info['nFolders'],
                            'Number of Slides': folder_info['nItems'],
                            'Last Updated': f['updated']
                        }
                    )


        return folder_slides, folder_folders

    def gen_collections_dataframe(self):
        """Generating dataframe containing current collections

        :return: Dataframe with each Collection
        :rtype: pd.DataFrame
        """
        collections_info = []
        collections = self.handler.get_collections()
        for c in collections:
            folder_count = self.handler.get_path_info(path = f'/collection/{c["name"]}')
            collections_info.append({
                'Name': c['name'],
                'ID': c['_id'],
                'Number of Folders': folder_count['nFolders'],
                'Last Updated': folder_count['updated']
            } | c['meta'])
            
        collections_df = pd.DataFrame.from_records(collections_info)

        return collections_df

    def extract_path_parts(self, current_parts:Union[list,dict], search_key: list = ['props','children'])->tuple:
        """Recursively extract pieces of folder paths stored as clickable components.

        :param current_parts: list or dictionary containing html.A or dbc.Stack of html.A components.
        :type current_parts: Union[list,dict]
        :param search_key: Property keys to search for in nested dicts, defaults to ['props','children']
        :type search_key: list, optional
        :return: Tuple containing all the parts of the folder path
        :rtype: tuple
        """
        path_pieces = ()
        if type(current_parts)==list:
            for c in current_parts:
                if type(c)==dict:
                    for key,value in c.items():
                        if key in search_key:
                            if type(value)==str:
                                path_pieces += (value,)
                            elif type(value) in [list,dict]:
                                path_pieces += self.extract_path_parts(value)
                elif type(c)==list:
                    path_pieces += self.extract_path_parts(c)
        elif type(current_parts)==dict:
            for key,value in current_parts.items():
                if key in search_key:
                    if type(value)==str:
                        path_pieces += (value,)
                    elif type(value) in [list,dict]:
                        path_pieces += self.extract_path_parts(value)

        return path_pieces

    def get_clicked_part(self, current_parts: Union[list,dict])->list:
        """Get the "n_clicks" value for components which have "id". If they have "id" but not "n_clicks", assign 0

        :param current_parts: Either a list or dictionary containing components
        :type current_parts: Union[list,dict]
        :return: List of values corresponding to "n_clicks" 
        :rtype: list
        """
        
        n_clicks_list = []
        if type(current_parts)==list:
            for c in current_parts:
                if type(c)==dict:
                    if 'id' in list(c.keys()):
                        if 'n_clicks' in list(c.keys()):
                            n_clicks_list.append(c['n_clicks'])
                        else:
                            n_clicks_list.append(0)
                    else:
                        for key,value in c.items():
                            if key=='props':
                                n_clicks_list += self.get_clicked_part(value)
                            elif key=='n_clicks':
                                n_clicks_list.append(value)
                elif type(c)==list:
                    n_clicks_list += self.get_clicked_part(c)
        elif type(current_parts)==dict:
            if 'id' in list(current_parts.keys()):
                if 'n_clicks' in list(current_parts.keys()):
                    n_clicks_list.append(current_parts['n_clicks'])
                else:
                    n_clicks_list.append(0)
            else:
                for key,value in current_parts.items():
                    if type(value)==dict:
                        if key=='props':
                            n_clicks_list += self.get_clicked_part(value)
                        elif key=='n_clicks':
                            n_clicks_list.append(value)
                    elif type(value)==list:
                        n_clicks_list += self.get_clicked_part(value)

        return n_clicks_list

    def gen_metadata_table(self, required_metadata: list, upload_items: list):
        """Generates metadata tables that the user can modify to add metadata to an uploaded item.

        :param required_metadata: List of required metadata items
        :type required_metadata: list
        :param upload_items: List of the names of uploaded items which can be used as targets of metadata values
        :type upload_items: list
        """
        dict_items = [i for i in required_metadata if type(i)==dict]
        dropdown_rows = [i for i in dict_items if 'values' in i]
        free_rows = [{'name':i, 'item': '', 'required': False} for i in required_metadata if type(i)==str]
        free_rows += [{'name': i['name'], 'item': i['item'], 'required': i['required']} for i in required_metadata if not 'values' in i]
        
        table_list = []
        for m_idx,m in enumerate([dropdown_rows,free_rows]):
            # Skip generating this table if now rows are present
            if len(m)==0:
                continue

            m_df = pd.DataFrame.from_records([
                {'Target': i['item'], 'Key': i['name'],'Value': '','row_id': idx}
                for idx,i in enumerate(m)
            ])
            required_rows = [r_idx for r_idx,r in enumerate(m) if r['required']]

            dropdown_conditional = [
                {
                    'if': {
                        'column_id': 'Target',
                        'filter_query': '{row_id} eq '+str(b)
                    },
                    'options': [
                        {'label': c, 'value': c}
                        for c in upload_items
                    ]
                }
                for b in range(len(m)) if m[b]['item']==''
            ]

            if m_idx==0:
                dropdown_conditional += [
                    {
                        'if': {
                            'column_id': 'Value',
                            'filter_query': '{row_id} eq '+str(k)
                        },
                        'options': [
                            {'label': l, 'value': l}
                            for l in m[k]['values']
                        ]
                    }
                    for k in range(len(m))
                ]

            metadata_table = dash_table.DataTable(
                id = {'type': f'{self.component_prefix}-dsa-uploader-metadata-table','index': m_idx},
                data = m_df.to_dict('records'),
                columns = [
                    {'id': 'Target', 'name': 'Target','presentation': 'dropdown'},
                    {'id': 'Key','name': 'Key'},
                    {'id': 'Value','name': 'Value','presentation': 'dropdown'} if m_idx==0 else {'id': 'Value','name':'Value'}
                ],
                editable = True,
                style_data_conditional = [
                    {
                        'if': {
                            'row_index': required_rows
                        },
                        'border': '2px solid rgb(255,0,0)'
                    }
                ],
                dropdown_conditional = dropdown_conditional,
                page_current = 0,
                page_size = 10,
                tooltip_data = [
                    {
                        column:{'value':str(value), 'type':'markdown'}
                        for column,value in row.items()
                    } for row in m_df.to_dict('records')
                ],
                tooltip_duration = None,
                css=[{"selector": ".Select-menu-outer", "rule": "display: block !important"}]
            )
            table_list.append(metadata_table)

        # Creating a "Custom Metadata" table where you can add rows
        custom_metadata_table = dbc.Stack([
            dash_table.DataTable(
                id = {'type': f'{self.component_prefix}-dsa-uploader-metadata-table','index': m_idx+1},
                data = [
                    {'Target': '', 'Key': '', 'Value': ''}
                ], 
                columns = [
                    {'id': 'Target', 'name': 'Target','presentation': 'dropdown'},
                    {'id': 'Key', 'name': 'Key'},
                    {'id': 'Value', 'name': 'Value'}
                ],
                dropdown = {
                    'Target': {
                        'options': [
                            {'label': w, 'value': w}
                            for w in upload_items
                        ]
                    }
                },
                editable = True,
                row_deletable = True,
                page_current = 0,
                page_size = 10,
                #tooltip_data = [
                #    {
                #        column:{'value':str(value), 'type':'markdown'}
                #        for column,value in row.items()
                #    } for row in m_df.to_dict('records')
                #],
                #tooltip_duration = None,
                css=[{"selector": ".Select-menu-outer", "rule": "display: block !important"}]
            ),
            dbc.Button(
                'Add Row',
                className = 'd-grid col-12 mx-auto',
                color = 'success',
                n_clicks = 0,
                id = {'type': f'{self.component_prefix}-dsa-uploader-metadata-table-add-row','index': m_idx+1}
            )
        ],direction='vertical',gap=2)

        table_list.append(custom_metadata_table)
        

        return table_list
    
    def populate_folder_div(self, collection_clicked, user_clicked, folder_table_rows, folder_crumb, back_clicked, folder_table_data, folder_crumb_parent,session_data):
        """Generate the collection/user folder div.

        :param collection_clicked: Whether "Collection" was clicked
        :type collection_clicked: list
        :param user_clicked: Whether "User" was clicked
        :type user_clicked: list
        :param folder_table_rows: Which rows in the folder table were clicked (all set to multi_row=False).
        :type folder_table_rows: list
        :param folder_crumb: Whether a part of the file path was clicked.
        :type folder_crumb: list
        :param back_clicked: Whether the back arrow as clicked.
        :type back_clicked: list
        :param folder_table_data: The current row data in the folder table.
        :type folder_table_data: list
        :param folder_crumb_parent: The parent container of all the folder path parts.
        :type folder_crumb_parent: list
        :param session_data: Current visualization session data
        :type session_data: list
        """

        path_parts = list(self.extract_path_parts(get_pattern_matching_value(folder_crumb_parent)))
        session_data = json.loads(session_data)
        folder_table_rows = get_pattern_matching_value(folder_table_rows)
        folder_table_data = get_pattern_matching_value(folder_table_data)

        if not ctx.triggered_id:
            raise exceptions.PreventUpdate

        if 'dsa-uploader-collection-button' in ctx.triggered_id['type']:
            
            if not any([i['value'] for i in ctx.triggered]):
                raise exceptions.PreventUpdate

            collection_df = self.gen_collections_dataframe()
            folder_table_div = html.Div([
                html.Div(
                    children = self.make_selectable_dash_table(
                        dataframe = collection_df,
                        id = {'type': f'{self.component_prefix}-dsa-uploader-folder-table','index': 0},
                        multi_row = False,
                        selected_rows = []
                    ),
                    id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                )
            ])

            path_parts = ['/collection/']
        
        elif 'dsa-uploader-user-folder-button' in ctx.triggered_id['type']:
            if not any([i['value'] for i in ctx.triggered]):
                raise exceptions.PreventUpdate
            
            user_folder_df = pd.DataFrame.from_records([
                {
                    'Name': i
                }
                for i in ['Private','Public']
            ])
            folder_table_div = html.Div([
                html.Div(
                    children = self.make_selectable_dash_table(
                        dataframe=user_folder_df,
                        id = {'type': f'{self.component_prefix}-dsa-uploader-folder-table','index': 0},
                        multi_row = False,
                        selected_rows = []
                    ),
                    id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                )
            ])

            path_parts = ['/user/',f'{session_data["current_user"]["login"]}/']

        elif 'dsa-uploader-folder-table' in ctx.triggered_id['type']:
            
            if not any([i['value'] for i in ctx.triggered]):
                raise exceptions.PreventUpdate
            
            path_parts = path_parts+[folder_table_data[folder_table_rows[0]]['Name']+'/']

            folder_info = self.handler.get_path_info(
                    path = ''.join(path_parts)[:-1]
                )

            # Don't need to know the slides in that folder
            _, folder_folders = self.organize_folder_contents(
                folder_info=folder_info
            )

            if len(folder_folders)>0:
                folder_table_div = html.Div([
                    html.Div(
                        children = self.make_selectable_dash_table(
                            dataframe = pd.DataFrame.from_records(folder_folders),
                            id = {'type': f'{self.component_prefix}-dsa-uploader-folder-table','index': 0},
                            multi_row = False,
                            selected_rows=[]
                        ),
                        id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                    ),
                    html.Hr(),
                        dbc.Stack([
                            dbc.Button(
                                'Create New Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'primary',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-button','index': 0}
                            ),
                            dbc.Button(
                                'Upload to this Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'success',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-select-folder','index': 0}
                            )
                        ],direction='horizontal',gap = 1)
                ])
            else:
                folder_table_div = html.Div([
                    html.Div(
                        children = dbc.Alert('This folder contains no more folders',color='warning'),
                        id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                    ),
                    html.Hr(),
                        dbc.Stack([
                            dbc.Button(
                                'Create New Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'primary',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-button','index': 0}
                            ),
                            dbc.Button(
                                'Upload to this Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'success',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-select-folder','index': 0}
                            )
                        ],direction='horizontal',gap = 1)
                ])
        
        elif 'dsa-uploader-folder-crumb' in ctx.triggered_id['type']:

            n_clicks = self.get_clicked_part(folder_crumb_parent)
            n_click_idx = np.argmax(n_clicks)

            if sum(n_clicks)==0:
                raise exceptions.PreventUpdate

            if n_click_idx==len(path_parts)-1:
                folder_table_div = html.Div([
                    html.Div(
                        children = self.make_selectable_dash_table(
                            dataframe = pd.DataFrame.from_records(folder_table_data),
                            id = {'type': f'{self.component_prefix}-dsa-uploader-folder-table','index': 0},
                            multi_row = False,
                            selected_rows = []
                        ),
                        id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                    ) if not folder_table_data is None else 
                    html.Div(
                        children = dbc.Alert('This folder contains no more folders',color='warning'),
                        id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                    ),
                    html.Hr(),
                        dbc.Stack([
                            dbc.Button(
                                'Create New Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'primary',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-button','index': 0}
                            ),
                            dbc.Button(
                                'Upload to this Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'success',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-select-folder','index': 0}
                            )
                        ],direction='horizontal',gap = 1)
                ])

                path_parts = path_parts[:-1]

            elif n_click_idx>0:

                path_parts = path_parts[:n_click_idx+1]
                folder_info = self.handler.get_path_info(
                    path = ''.join(path_parts)[:-1]
                )

                if not path_parts==['/user/',session_data['current_user']['login']+'/']:
                    # Don't need to know the slides in that folder
                    _, folder_folders = self.organize_folder_contents(
                        folder_info=folder_info
                    )
                else:
                    folder_folders = [
                        {
                            'Name': i
                        }
                        for i in ['Private','Public']
                    ]


                if len(folder_folders)>0:
                    folder_table_div = html.Div([
                        html.Div(
                            children = self.make_selectable_dash_table(
                                dataframe = pd.DataFrame.from_records(folder_folders),
                                id = {'type': f'{self.component_prefix}-dsa-uploader-folder-table','index': 0},
                                multi_row = False,
                                selected_rows=[]
                            ),
                            id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                        ),
                        html.Hr(),
                        dbc.Stack([
                            dbc.Button(
                                'Create New Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'primary',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-button','index': 0}
                            ),
                            dbc.Button(
                                'Upload to this Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'success',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-select-folder','index': 0}
                            )
                        ],direction='horizontal',gap = 1)
                    ])
                else:
                    folder_table_div = html.Div([
                        html.Div(
                            children = dbc.Alert('This folder contains no more folders',color='warning'),
                            id = {'type':f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                        ),
                        html.Hr(),
                        dbc.Stack([
                            dbc.Button(
                                'Create New Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'primary',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-button','index': 0}
                            ),
                            dbc.Button(
                                'Upload to this Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'success',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-select-folder','index': 0}
                            )
                        ],direction='horizontal',gap = 1)
                    ])

            elif n_click_idx==0:
                path_parts = path_parts[0]

                if path_parts == '/collection/':
                    folder_folders = self.gen_collections_dataframe().to_dict('records')
                elif path_parts =='/user/':
                    folder_folders = [
                        {
                            'Name': i['login']
                        }
                        for i in self.handler.gc.get(f'/user?token={session_data["current_user"]["token"]}')
                    ]

                if len(folder_folders)>0:
                    folder_table_div = html.Div([
                        html.Div(
                            children = self.make_selectable_dash_table(
                                dataframe = pd.DataFrame.from_records(folder_folders),
                                id = {'type': f'{self.component_prefix}-dsa-uploader-folder-table','index': 0},
                                multi_row = False,
                                selected_rows=[]
                            ),
                            id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                        ),
                        html.Hr(),
                        dbc.Stack([
                            dbc.Button(
                                'Create New Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'primary',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-button','index': 0}
                            ),
                            dbc.Button(
                                'Upload to this Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'success',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-select-folder','index': 0}
                            )
                        ],direction='horizontal',gap = 1)
                    ])
                else:
                    folder_table_div = html.Div([
                        html.Div(
                            children = dbc.Alert('This folder contains no more folders',color='warning'),
                            id = {'type': f'{self.component_prefix}-dsa-uploader-folder-div','index': 0}
                        ),
                        html.Hr(),
                        dbc.Stack([
                            dbc.Button(
                                'Create New Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'primary',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-button','index': 0}
                            ),
                            dbc.Button(
                                'Upload to this Folder',
                                className = 'd-grid col-6 mx-auto',
                                color = 'success',
                                n_clicks = 0,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-select-folder','index': 0}
                            )
                        ],direction='horizontal',gap = 1)
                    ])

        elif 'dsa-uploader-folder-back-icon' in ctx.triggered_id['type']:

            collection_or_user_div_children = [
                html.Div(
                    id = {'type': f'{self.component_prefix}-dsa-uploader-folder-nav-parent','index': 0},
                    children = []
                ),
                dbc.Stack([
                    dbc.Button(
                        'Collection',
                        n_clicks = 0,
                        className = 'd-grid col-6 mx-auto',
                        color = 'primary',
                        id = {'type': f'{self.component_prefix}-dsa-uploader-collection-button','index': 0}
                    ),
                    dbc.Button(
                        'User Folder',
                        n_clicks = 0,
                        className = 'd-grid col-6 mx-auto',
                        color = 'secondary',
                        id = {'type': f'{self.component_prefix}-dsa-uploader-user-folder-button','index': 0}
                    )
                ],direction = 'horizontal',gap = 3)
            ]

            return [collection_or_user_div_children], [html.Div()], [html.Div()]

        new_crumbs = []
        for i in path_parts:
            new_crumbs.append(
                html.A(
                    i,
                    id = {'type': f'{self.component_prefix}-dsa-uploader-folder-crumb','index': 0},
                    style = {'color': 'rgb(0,0,255)'}
                )
            )

        collection_or_user_div_children = html.Div([
            html.Div(
                html.A(
                    children = dbc.Stack([
                        html.P(
                            html.I(
                                className = 'fa-solid fa-arrow-left',
                                style = {'marginRight': '2px'}
                            )
                        ),
                        html.P(
                            'Back'
                        )
                    ],direction = 'horizontal'),
                    n_clicks = 0,
                    id = {'type': f'{self.component_prefix}-dsa-uploader-folder-back-icon','index': 0}
                ),
                id = {'type': f'{self.component_prefix}-dsa-uploader-folder-back-icon-div','index': 0}
            ),
            html.H5(
                children = [
                    dbc.Stack(new_crumbs,direction='horizontal',gap=1)
                ],
                id = {'type': f'{self.component_prefix}-dsa-uploader-folder-nav-parent','index': 0},
                style = {'textTransform': 'none','display': 'inline'}
            ),
            html.Hr(),
            folder_table_div
        ])

        # Clearing the new folder div
        new_folder_div = html.Div()

        return [collection_or_user_div_children], [new_crumbs], [new_folder_div]
    
    def populate_new_folder(self, create_clicked, submit_clicked, cancel_clicked, new_folder_name, parent_path, session_data):
        """Callback for creating a new folder at a specific location.

        :param create_clicked: Whether Create Folder was clicked
        :type create_clicked: list
        :param submit_clicked: Whether Submit folder was clicked
        :type submit_clicked: list 
        :param cancel_clicked: Whether Cancel was clicked
        :type cancel_clicked: list                 
        :param new_folder_name: Name of new folder
        :type new_folder_name: list
        :param parent_path: Parent of folder path parts
        :type parent_path: list
        :param session_data: Current Visualization Session data
        :type session_data: list
        """

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        session_data = json.loads(session_data)

        if 'dsa-uploader-new-folder-button' in ctx.triggered_id['type']:

            folder_table_div = no_update
            new_path_parts = no_update
            new_folder_disable = True

            new_folder_div = html.Div([
                dbc.InputGroup([
                    dbc.InputGroupText(
                        'Folder Name: '
                    ),
                    dbc.Input(
                        id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-input','index': 0},
                        placeholder = 'fusion-tools Upload',
                        type = 'text',
                        required = True,
                        value = [],
                        maxLength = 1000,
                    ),
                    dbc.Button(
                        'Create Folder!',
                        color = 'primary',
                        n_clicks = 0,
                        id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-submit-button','index': 0}
                    ),
                    dbc.Button(
                        'Cancel',
                        color = 'danger',
                        n_clicks = 0,
                        id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-cancel-button','index': 0}
                    )
                ])
            ],style = {'marginTop':'10px'})

            error_div = []

        elif 'dsa-uploader-new-folder-submit-button' in ctx.triggered_id['type']:

            folder_path = list(self.extract_path_parts(get_pattern_matching_value(parent_path)))
            folder_name = get_pattern_matching_value(new_folder_name)

            try:
                new_folder_info = self.handler.create_user_folder(
                    parent_path = ''.join(folder_path)[:-1],
                    folder_name = folder_name,
                    user_token=session_data['current_user']['token']
                )
            except Exception as e:
                print('Some error encountered in creating the folder')
                print(f'Status Code: {e.status_code}')
                print(e.json())
                new_folder_info = {'ok':False}
                error_div = dbc.Alert(f'Error creating folder at: {"".join(folder_path)+folder_name}',color='danger')

            if new_folder_info.ok and new_folder_info.status_code==200:
                error_div = dbc.Alert(f'Success!',color = 'success')
                new_folder_disable = False

                new_path_parts = []
                for i in folder_path+[folder_name]:
                    new_path_parts.append(
                        html.A(
                            i,
                            id = {'type': f'{self.component_prefix}-dsa-uploader-folder-crumb','index': 0},
                            style = {'color': 'rgb(0,0,255)'}
                        )
                    )

                new_path_parts = dbc.Stack(new_path_parts,direction='horizontal',gap=1)

                new_folder_div = html.Div([
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'Folder Name: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-input','index': 0},
                            placeholder = 'fusion-tools Upload',
                            type = 'text',
                            required = True,
                            value = folder_name,
                            maxLength = 1000,
                            disabled=True
                        ),
                        dbc.Button(
                            'Create Folder!',
                            color = 'success',
                            disabled = True,
                            n_clicks = 0,
                            id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-submit-button','index': 0}
                        ),
                        dbc.Button(
                            'Cancel',
                            color = 'danger',
                            disabled = True,
                            n_clicks = 0,
                            id = {'type': f'{self.component_prefix}-dsa-uploader-new-folder-cancel-button','index': 0}
                        )
                    ])
                ],style = {'marginTop':'10px'})

                folder_table_div = html.Div([
                    dbc.Alert('This folder contains no more folders',color = 'warning')
                ])

            else:
                new_path_parts = no_update
                new_folder_div = no_update
                folder_table_div = no_update
                new_folder_disable = True

                error_div = dbc.Alert(f'Error creating folder at: {"".join(folder_path)+folder_name}',color='danger')

        elif 'dsa-uploader-new-folder-cancel-button' in ctx.triggered_id['type']:
            new_folder_div = html.Div()
            error_div = []
            folder_table_div = no_update
            new_path_parts = no_update
            new_folder_disable = False


        return [new_folder_div], [error_div], [folder_table_div], [new_path_parts], [new_folder_disable]

    def populate_upload_type(self, select_clicked,path_parts):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        upload_div_contents = html.Div([
            dbc.Row([
                dbc.Col([
                    dbc.Card([
                        dbc.CardHeader('Available Upload Types'),
                        dbc.CardBody([
                            html.Div([
                                #html.P('Different data types require different sets and types of files in order to have sufficient information for visualization and analysis'),
                                #html.B(),
                                html.H5('Select the type of data you would like to upload from the menu below'),
                                html.Hr(),
                                dcc.Dropdown(
                                    options = [
                                        {
                                            'label': i.name, 'value': i.name
                                        }
                                        for i in self.dsa_upload_types
                                    ],
                                    value = [],
                                    placeholder='Selected Upload Type',
                                    id = {'type': f'{self.component_prefix}-dsa-uploader-upload-type-drop','index': 0}
                                ),
                                html.B(),
                                html.Div(
                                    id = {'type': f'{self.component_prefix}-dsa-uploader-upload-type-description-div','index': 0},
                                    children = []
                                )
                            ],style = {'height': '30vh','maxHeight': '40vh','overflow': 'scroll'})
                        ])
                    ])
                ],md = 4),
                dbc.Col([
                    dbc.Card([
                        dbc.CardHeader('Upload Files'),
                        dbc.CardBody([
                            dcc.Store(
                                id = {'type': f'{self.component_prefix}-dsa-uploader-upload-files-store','index': 0},
                                storage_type = 'memory',
                                data = json.dumps({'uploaded_files': []})
                            ),
                            html.Div(
                                id = {'type': f'{self.component_prefix}-dsa-uploader-upload-type-upload-files-div','index': 0},
                                children = [],
                                style = {'maxHeight': '40vh','overflow': 'scroll'}
                            )
                        ])
                    ])
                ])
            ])
        ])

        path_parts = list(self.extract_path_parts(get_pattern_matching_value(path_parts)))
        disabled_path_parts = []
        for p in path_parts:
            disabled_path_parts.append(
                html.P(p)
            )
        
        disabled_path_parts = dbc.Stack(
            disabled_path_parts,
            direction = 'horizontal'
        )

        back_icon_children = dbc.Stack([
                html.P(
                    html.I(
                        className = 'fa-solid fa-arrow-left',
                        style = {'marginRight': '2px'}
                    )
                ),
                html.P(
                    'Back'
                )
            ],direction = 'horizontal'
        )

        new_folder_disable = True
        select_folder_disable = True

        return [upload_div_contents], [disabled_path_parts], [new_folder_disable], [select_folder_disable], [back_icon_children]

    def create_upload_component(self, file_info, user_info, idx):
        
        # "item" type uploads have a parentId = the Id of the folder they are being uploaded to
        # "file" and "annotation" type uploads have a parentId = the item that they are being added to
        # if the parent "item" for a "file" or "annotation" does not exist yet, then the uploader should be disabled

        upload_id_dict = {
            'api_url': self.handler.girderApiUrl,
            'token': user_info['token'],
            'fusion_upload_name': file_info['name'],
            'fusion_upload_type': file_info['type'],
            'parentType': file_info['parentType'],
            'parentId': file_info['parentId']
        }

        upload_component = du.Upload(
            id = {'type': f'{self.component_prefix}-dsa-uploader-file-upload','index': idx},
            max_file_size = MAX_UPLOAD_SIZE,
            chunk_size = MIN_UPLOAD_SIZE,
            filetypes = file_info['accepted_types'],
            cancel_button = True,
            pause_button = True,
            upload_id = json.dumps(upload_id_dict),
            disabled = False
        )

        return upload_component

    def make_file_uploads(self, upload_type_value, upload_folder_path, session_data):
        """Making the file upload components for the selected UploadType

        :param upload_type_value: Selected UploadType from the dropdown menu
        :type upload_type_value: list
        :param upload_folder_path: Folder path to upload to
        :type upload_folder_path: list
        :param session_data: Current visualization session data
        :type session_data: str
        """

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        upload_type_value = get_pattern_matching_value(upload_type_value)

        upload_path_parts = self.extract_path_parts(get_pattern_matching_value(upload_folder_path))
        folder_info = self.handler.get_path_info(
            path = ''.join(upload_path_parts)[:-1]
        )

        selected_upload_type = self.dsa_upload_types[[i.name for i in self.dsa_upload_types].index(upload_type_value)]

        file_uploads = html.Div([
            dbc.Stack([
                html.Div([
                    html.H5(f'{f["name"]}',style={'textTransform':'none'}),
                    html.Div(
                        self.create_upload_component(
                            file_info = f | {'parentId': folder_info["_id"],'parentType': 'folder', 'fusion_upload_name': f['name'], 'fusion_upload_type': f['type']},
                            user_info = session_data['current_user'],
                            idx = f_idx
                        ) if f['type']=='item' else 
                        dbc.Alert(f'This upload will be created when: {f["parent"]} is uploaded',color='warning'),
                        id = {'type': f'{self.component_prefix}-dsa-uploader-file-upload-div','index': f_idx},
                        style = {'width': '100%'}
                    ),
                    dbc.Tooltip(
                        target = {'type': f'{self.component_prefix}-dsa-uploader-file-upload-div','index': f_idx},
                        placement='top',
                        children = ','.join(f['accepted_types'])
                    ),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-uploader-file-upload-status-div','index': f_idx},
                        children = []
                    ),
                    html.P(f['description'] if 'description' in f else '')
                ])
                for f_idx,f in enumerate(selected_upload_type.input_files)
            ]),
            dbc.Row([
                dbc.Col(
                    html.Div(
                        dbc.Button(
                            'Done!',
                            className = 'd-grid col-4 mx-auto',
                            n_clicks = 0,
                            color = 'success',
                            disabled = True,
                            id = {'type': f'{self.component_prefix}-dsa-uploader-file-upload-done-button','index': 0}
                        )
                    )
                )
            ],align='right')
        ])

        upload_type_description = html.P(selected_upload_type.description)

        return [upload_type_description], [file_uploads]

    def enable_upload_done(self, uploads_complete, upload_type, current_filenames,session_data,upload_file_data):
        """Enabling the "Done" button when all required uploads are uploaded

        :param uploads_complete: Current uploadComplete flags from active UploadComponents
        :type uploads_complete: list
        :param upload_type: Selected type of upload
        :type upload_type: list
        """
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        upload_file_data = json.loads(get_pattern_matching_value(upload_file_data))
        upload_type = get_pattern_matching_value(upload_type)
        selected_upload_type = self.dsa_upload_types[[i.name for i in self.dsa_upload_types].index(upload_type)]

        # Getting the filenames (json.dumps(post_response))
        current_filenames = [i for i in current_filenames if not i is None]
        if len(current_filenames)>0:
            current_filenames = current_filenames[0]
        else:
            current_filenames = None
        completed_filenames = []
        if current_filenames is None:
            raise exceptions.PreventUpdate
        
        for c in current_filenames:
            if not c is None:
                #TODO: "fileNames" is a list of files by default so in theory this could be expanded to accept multiple
                # uploads to the same container
                if type(c)==str:
                    completed_filenames.append(json.loads(c))
                elif type(c)==list:
                    completed_filenames.append(json.loads(c[0]))

        upload_file_data['uploaded_files'].extend(completed_filenames)
        input_file_types = [i['fusion_upload_name'] for i in completed_filenames][0]

        completed_upload_file = selected_upload_type.input_files[[i['name'] for i in selected_upload_type.input_files].index(input_file_types)]
        child_files = [
            [idx,i] for idx,i in enumerate(selected_upload_type.input_files) if i['parent']==input_file_types
        ]

        upload_div_children = [no_update]*len(ctx.outputs_list[0])
        for c in child_files:
            file_info = {
                'api_url': self.handler.girderApiUrl,
                'token': session_data['current_user']['token'],
                'fusion_upload_name': c[1]['name'],
                'fusion_upload_type': c[1]['type'],
                'parentType': 'item',
                'parentId': completed_filenames[0]['itemId']
            }

            upload_div_children[c[0]] = self.create_upload_component(
                file_info = c[1] | file_info,
                user_info = session_data['current_user'],
                idx = c[0]
            )

        if 'name' in completed_filenames[0]:
            upload_div_children[ctx.triggered_id['index']] = dbc.Alert(f'Success! {completed_filenames[0]["name"]}', color = 'success')
        elif 'n_annotations' in completed_filenames[0]:
            upload_div_children[ctx.triggered_id['index']] = dbc.Alert(f'Success! {completed_filenames[0]["n_annotations"]} Annotations Added!',color = 'success')
        else:
            upload_div_children[ctx.triggered_id['index']] = dbc.Alert(f'Success!', color = 'success')

        # Checking if all required uploads are done:
        required_files = [i['name'] for i in selected_upload_type.input_files if i['required']]
        uploaded_files = [i['fusion_upload_name'] for i in upload_file_data['uploaded_files']]
        
        if len(set(required_files).difference(uploaded_files))==0:
            done_disabled = [False]
        else:
            done_disabled = [True]
        
        # Disabling upload type drop
        upload_type_disabled = [True]

        return upload_div_children, done_disabled, upload_type_disabled, [json.dumps(upload_file_data)]

    def populate_processing_plugins(self, done_clicked,upload_type,path_parts,upload_files_data, session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        upload_type = get_pattern_matching_value(upload_type)
        selected_upload_type = self.dsa_upload_types[[i.name for i in self.dsa_upload_types].index(upload_type)]

        session_data = json.loads(session_data)
        upload_files_data = json.loads(get_pattern_matching_value(upload_files_data))
        # Modifying session data to include information on uploaded files (used by plugin inputs handler)
        session_data['uploaded_files'] = upload_files_data['uploaded_files']

        uploaded_items = [i['name'] for i in selected_upload_type.input_files if i['type']=='item']
        metadata_table_list = self.gen_metadata_table(selected_upload_type.required_metadata, uploaded_items)
        any_required = [i for i in selected_upload_type.required_metadata if type(i)==dict]
        any_required = any([i['required'] for i in any_required if 'required' in i])

        # Getting input components for the processing plugins
        processing_plugin_children = self.plugin_inputs_handler.update_layout(
            session_data=session_data,
            use_prefix=True,
            plugin_groups=selected_upload_type.processing_plugins
        )

        processing_plugin_div = html.Div([
            dbc.Row([
                dbc.Col([
                    dbc.Card([
                        dbc.CardHeader('Required Metadata'),
                        dbc.CardBody([
                            html.P('Add any required metadata below: '),
                            html.Hr(),
                            dbc.Stack(metadata_table_list,direction='vertical',style = {'marginBottom':'5px'}),
                            html.B(),
                            dbc.Button(
                                'Submit Metadata',
                                className = 'd-grid col-12 mx-auto',
                                n_clicks = 0,
                                color = 'success',
                                disabled = any_required,
                                id = {'type': f'{self.component_prefix}-dsa-uploader-metadata-submit-button','index': 0}
                            ),
                            html.Div(
                                id = {'type': f'{self.component_prefix}-dsa-uploader-metadata-submit-status-div','index': 0},
                                children = []
                            )
                        ])
                    ])
                ],md = 6),
                dbc.Col([
                    dbc.Card([
                        dbc.CardHeader('Processing Plugins'),
                        dbc.CardBody([
                            html.Div(
                                children = processing_plugin_children
                            ),
                            html.Div(
                                id = {'type': f'{self.component_prefix}-dsa-uploader-all-plugins-run-status','index': 0},
                                children = []
                            )
                        ])
                    ])
                ],md = 6)
            ])
        ])

        upload_type_disable = True
        file_upload_divs = [dbc.Alert('Done Uploading',color = 'secondary')]*len(ctx.outputs_list[2])
        done_button_disable = True

        return [processing_plugin_div], [upload_type_disable], file_upload_divs, [done_button_disable]
    
    def enable_submit_metadata(self, all_table_data, upload_type):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        upload_type = get_pattern_matching_value(upload_type)
        selected_upload_type = self.dsa_upload_types[[i.name for i in self.dsa_upload_types].index(upload_type)]

        # Getting which metadata fields are required
        required_metadata = [i for i in selected_upload_type.required_metadata if type(i)==dict]
        required_metadata= [i['name'] for i in required_metadata if i['required']]

        req_meta_check = []
        for t in all_table_data:
            for d in t:
                if d['Key'] in required_metadata:
                    req_meta_check.append(not d['Value']=='' and not d['Target']=='')
        
        submit_disable = not all(req_meta_check)

        return [submit_disable]

    def add_row_custom_metadata(self, clicked, custom_metadata):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        # Just appending a new blank row to the current data
        custom_metadata.append(
            {'Target': '', 'Key': '', 'Value': ''}
        )

        return custom_metadata

    def submit_metadata(self, clicked, upload_files_data, tables_data, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        upload_files_data = json.loads(get_pattern_matching_value(upload_files_data))['uploaded_files']
        upload_names = [i['fusion_upload_name'] for i in upload_files_data]

        target_dict = {}
        for t in tables_data:
            for row in t:
                if not row['Target']=='':
                    target_id = upload_files_data[upload_names.index(row['Target'])]['itemId']

                    if target_id in target_dict:
                        target_dict[target_id] = target_dict[target_id] | {row['Key']:row['Value']}
                    else:
                        target_dict[target_id] = {row['Key']:row['Value']}

        status_div = []
        for t in list(target_dict.keys()):
            success = self.handler.add_metadata(
                item = t,
                metadata = target_dict[t],
                user_token = session_data['current_user']['token']
            )
            
            if success:
                status_div.append(dbc.Alert(f'Metadata added to {t}!',color = 'success'))
            else:
                status_div.append(dbc.Alert(f'Error adding metadata to item: {t}',color='danger'))

        return status_div







"""

Components which deal with access of external APIs for data/embeddable widgets


"""

import os
import sys
import json
import numpy as np
import pandas as pd

from io import BytesIO
import requests

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, Patch, no_update, dash_table
from dash.dash_table.Format import Format, Scheme
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
#import dash_treeview_antd as dta
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform, BlockingCallbackTransform
from dash_extensions.javascript import Namespace, arrow_function

# fusion-tools imports
from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools import Tool

import time
from tqdm import tqdm


class HRAViewer(Tool):
    """HRAViewer Tool which enables hierarchy visualization for organs, cell types, biomarkers, and proteins in the Human Reference Atlas

    For more information on the Human Reference Atlas (HRA), see: https://humanatlas.io/. Thanks Bruce!

    :param Tool: General class for interactive components that visualize, edit, or perform analyses on data.
    :type Tool: None
    """
    def __init__(self):
        """Constructor method
        """
        
        super().__init__()

        self.hra_collection = "https://purl.humanatlas.io/collection/hra"

        try:
            self.asct_b_release = self.get_latest_asctb()

            self.organ_table_options = [
                {'label': f'{i} ASCT+B Table', 'value': i, 'disabled': False}
                for i in list(self.asct_b_release.keys())
            ]

            self.show_asct_b_error = False

        except Exception as e:
            print(f'Error initializing HRAViewer: {e}')
            self.asct_b_release = None
            self.organ_table_options = []

            self.show_asct_b_error = True


    def __str__(self):
        return "HRA Viewer"

    def fetch_json(self, purl_str: str):
        """Run GET request and return JSON data

        :param purl_str: String URL for request
        :type purl_str: str
        """

        response = requests.get(purl_str, headers = {'Accept': 'application/json'})
        response.raise_for_status()

        return response.json()
    
    def is_asctb_table(self, digital_object:str):
        """Check if digital object string is for an ASCT+B table

        :param digital_object: String corresponding to digital object
        :type digital_object: str
        """

        return (
            digital_object.startswith("https://purl.humanatlas.io/asct-b/")
            and "crosswalk" not in digital_object
        )

    def get_latest_asctb(self):
        """Get latest set of ASCT+B tables from Human Atlas collections
        """

        hra_collections = self.fetch_json(self.hra_collection)
        digital_objects = hra_collections["metadata"]["had_member"]

        iter_objects = sorted(filter(self.is_asctb_table,digital_objects))
        tables = {}
        with tqdm(iter_objects) as pbar:
            for purl in iter_objects:
                table_name = purl.split("/")[-2].replace('-','_')
                table_data = self.fetch_json(purl)
                table_attribution = self.fetch_json(purl.replace("purl","lod"))

                pbar.set_description(f'On: {table_name}')

                table_metadata = {
                    'version': table_attribution.get('version','v1.0'),
                    'creation_data': table_attribution.get('creation_date',''),
                    'doi': table_attribution.get('was_derived_from',{}).get('doi',''),
                    'citation': table_attribution.get('citiation',''),
                    'description': table_attribution.get('was_derived_from',{}).get('description',''),
                    'authors': [
                        {
                            'fullName': i.get('fullName',''),
                            'firstName': i.get('firstName',''),
                            'lastName': i.get('lastName',''),
                            'orcid': i.get('orcid','')
                        }
                        for i in table_attribution.get('was_derived_from',{}).get('creators',[])
                    ],
                    'reviewers': [
                        {
                            'fullName': i.get('fullName',''),
                            'firstName': i.get('firstName',''),
                            'lastName': i.get('lastName',''),
                            'orcid': i.get('orcid','')
                        }
                        for i in table_attribution.get('was_derived_from',{}).get('reviewers',[])
                    ]
                }

                table_rows = table_data['data']['asctb_record']

                organ_table = []
                for o_idx,o in enumerate(table_rows):
                    an_structs = o.get('anatomical_structure_list',[])
                    c_types = o.get('cell_type_list',[])
                    g_markers = o.get('gene_marker_list',[])
                    p_markers = o.get('protein_marker_list',[])

                    as_cols = [
                        {
                            f'{j.get("id","").split("#")[-1].split("-")[-1]}': j.get("ccf_pref_label","-"),
                            f'{j.get("id","").split("#")[-1].split("-")[-1]} ID': j.get("source_concept","-")
                        }
                        for j in an_structs
                    ]

                    ct_cols = [
                        {
                            f'{j.get("id","").split("#")[-1].split("-")[-1]}': j.get("ccf_pref_label","-"),
                            f'{j.get("id","").split("#")[-1].split("-")[-1]} ID': j.get("source_concept","-")
                        }
                        for j in c_types
                    ]

                    gm_cols = [
                        {
                            f'{j.get("id","").split("#")[-1].split("-")[-1]}': j.get("ccf_pref_label","-"),
                            f'{j.get("id","").split("#")[-1].split("-")[-1]} ID': j.get("source_concept","-")
                        }
                        for j in g_markers
                    ]

                    pm_cols = [
                        {
                            f'{j.get("id","").split("#")[-1].split("-")[-1]}': j.get("ccf_pref_label","-"),
                            f'{j.get("id","").split("#")[-1].split("-")[-1]} ID': j.get("source_concept","-")
                        }
                        for j in p_markers
                    ]

                    row_data = {}
                    for col in as_cols+ct_cols+gm_cols+pm_cols:
                        row_data = row_data | col

                    organ_table.append(row_data)


                tables[table_name] = {
                    'organ_table': organ_table,
                    'metadata': table_metadata
                }
                pbar.update(1)
            
        return tables

    def load(self, component_prefix: int):
        
        self.component_prefix = component_prefix
        self.title = 'HRA Viewer'
        self.blueprint = DashBlueprint(
            transforms = [
                PrefixIdTransform(prefix = f'{self.component_prefix}'),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

    def gen_layout(self, session_data: dict):
        """Generate layout for HRA Viewer component
        """

        layout = html.Div([
            dbc.Card([
                dbc.CardBody([
                    dbc.Row(
                        dbc.Col(html.H3('Human Reference Atlas (HRA) Viewers'))
                    ),
                    html.Hr(),
                    dbc.Row(
                        dbc.Col('Select one of the embedded components below or select an organ to view the ASCT+B table for that organ')
                    ),
                    html.Hr(),
                    html.Div(
                        id = {'type': 'hra-viewer-error-div','index': 0},
                        children = [
                            dbc.Alert(
                                color = 'danger',
                                children = [
                                    dbc.Row([
                                        dbc.Col([
                                            f'Error loading ASCT+B Tables'
                                        ],md = 8),
                                        dbc.Col([
                                            html.A(
                                                html.I(
                                                    className = 'fa-solid fa-rotate fa-2x',
                                                    n_clicks = 0,
                                                    id = {'type': 'hra-viewer-refresh-icon','index': 0}
                                                )
                                            ),
                                            dbc.Tooltip(
                                                target = {'type': 'hra-viewer-refresh-icon','index': 0},
                                                children = 'Click to re-try grabbing ASCT+B tables'
                                            )],
                                            md = 4
                                        )
                                    ])
                                ]
                            )
                        ] if self.show_asct_b_error else []
                    ),
                    dbc.Row([
                        dbc.Col([
                            dbc.Label('HRA View Select: ',html_for = {'type': 'hra-viewer-drop','index': 0})
                        ],md=2),
                        dbc.Col([
                            dcc.Dropdown(
                                options = [
                                    {'label': 'FTU Explorer','value': 'FTU Explorer','disabled': False},
                                ] + self.organ_table_options,
                                value = [],
                                multi = False,
                                id = {'type': 'hra-viewer-drop','index': 0}
                            )
                        ], md = 10)
                    ]),
                    dbc.Row([
                        dbc.Col([
                            html.Div(
                                id = {'type': 'hra-viewer-parent','index': 0},
                                children = [],
                            )
                        ])
                    ],align='center')
                ])
            ])
        ],style = {'width': '100%'})

        self.blueprint.layout = layout

    def get_callbacks(self):
        """Initializing callbacks and attaching to DashBlueprint
        """

        self.blueprint.callback(
            [
                Input({'type': 'hra-viewer-drop','index': ALL},'value')
            ],
            [
                Output({'type': 'hra-viewer-parent','index': ALL},'children')
            ]
        )(self.update_hra_viewer)

        self.blueprint.callback(
            [
                Input({'type': 'hra-viewer-refresh-icon','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'hra-viewer-error-div','index': ALL},'children'),
                Output({'type': 'hra-viewer-drop','index': ALL},'options')
            ]
        )(self.refresh_asctb_tables)

    def get_organ_table(self, organ:str):
        """Grabbing ASCT+B Table for a specific organ

        :param organ: Name of organ to get table for
        :type organ: str
        """
        
        if organ in self.asct_b_release:
            new_table = pd.DataFrame.from_records(self.asct_b_release[organ]['organ_table']).fillna('-')
            sorted_cols = sorted([i for i in new_table if 'AS' in i]) + sorted([i for i in new_table if 'CT' in i]) + sorted([i for i in new_table if 'BM' in i]) + sorted([i for i in new_table if 'P' in i])
            new_table = new_table.reindex(sorted_cols,axis=1)
            
            table_attribution = self.asct_b_release[organ]['metadata']

        else:
            new_table = None
            table_attribution = None

        return new_table, table_attribution

    def update_hra_viewer(self, viewer_drop_value):
        """Updating the HRAViewer component based on selected view

        :param viewer_drop_value: Selected component from dropdown (one of FTU Explorer or {organ} ASCT+B Table)
        :type viewer_drop_value: list
        """
        """
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        """

        viewer_drop_value = get_pattern_matching_value(viewer_drop_value)

        if viewer_drop_value is None:
            raise exceptions.PreventUpdate

        viewer_children = no_update
        if viewer_drop_value=='FTU Explorer':

            viewer_children = html.Iframe(
                srcDoc = '''
                    <!DOCTYPE html>
                    <html lang="en">
                    <head>
                        <meta charset="utf-8" />
                        <title>FTU Ui Small Web Component</title>
                        <meta name="viewport" content="width=device-width, initial-scale=1" />
                        <link href="https://fonts.googleapis.com/css2?family=Inter:wght@300;400;500&display=swap" rel="stylesheet" />
                        <link href="https://fonts.googleapis.com/icon?family=Material+Icons" rel="stylesheet" />
                        <link
                        rel="stylesheet"
                        href="https://fonts.googleapis.com/css2?family=Material+Symbols+Outlined:opsz,wght,FILL,GRAD@20..48,100..700,0..1,-50..200"
                        />
                        <link href="https://cdn.humanatlas.io/ui/ftu-ui-small-wc/styles.css" rel="stylesheet" />
                        <script src="https://cdn.humanatlas.io/ui/ftu-ui-small-wc/wc.js" defer></script>
                    </head>
                    <body style="margin: 0">
                        <hra-ftu-ui-small
                        base-href="https://cdn.humanatlas.io/ui/ftu-ui-small-wc/"
                        selected-illustration="https://purl.humanatlas.io/2d-ftu/kidney-renal-corpuscle"
                        datasets="assets/TEMP/ftu-datasets.jsonld"
                        summaries="assets/TEMP/ftu-cell-summaries.jsonld"
                        >
                        </hra-ftu-ui-small>
                    </body>
                    </html>
                ''',
                style = {
                    'height': '1000px','width': '100%','overflow': 'scroll'
                }
            )
        
        if not self.asct_b_release is None:
            if viewer_drop_value in self.asct_b_release:
                
                organ_table, table_attribution = self.get_organ_table(viewer_drop_value)
                organ_table[list(organ_table)] = organ_table[list(organ_table)].astype(str)

                
                if not organ_table is None and not table_attribution is None:
                    
                    reviewers_and_authors = [
                        dbc.Row([
                            dbc.Col([
                                dbc.Label(
                                    html.H4('Authors:'),
                                    html_for={'type': f'{self.component_prefix}-hra-viewer-authors','index': 0}
                                )
                            ])
                        ]),
                        dbc.Row([
                            dbc.Col([
                                html.Div(
                                    dmc.AvatarGroup(
                                        id = {'type': f'{self.component_prefix}-hra-viewer-authors','index': 0},
                                        children = [
                                            html.A(
                                                dmc.Tooltip(
                                                    dmc.Avatar(
                                                        ''.join([n[0] for n in author.get('fullName').split()]),
                                                        radius = 'xl',
                                                        size = 'lg',
                                                        color = f'rgb({np.random.randint(0,255)},{np.random.randint(0,255)},{np.random.randint(0,255)})'
                                                    ),
                                                    label = author.get('fullName'),
                                                    position = 'bottom'
                                                ),
                                                href = f'https://orcid.org/{author.get("orcid")}',
                                                target = '_blank'
                                            )
                                            for author in table_attribution['authors']
                                        ]
                                    )
                                )
                            ])
                        ],style={'marginBottom': '5px'},align='center'),
                        dbc.Row([
                            dbc.Col(
                                dbc.Label(
                                    html.H6(
                                        'Reviewers: '
                                    ),
                                    html_for={'type': f'{self.component_prefix}-hra-viewer-reviewers','index': 0}
                                )
                            )
                        ]),
                        dbc.Row([
                            dbc.Col([
                                html.Div(
                                    dmc.AvatarGroup(
                                        id = {'type': f'{self.component_prefix}-hra-viewer-reviewers','index': 0},
                                        children = [
                                            html.A(
                                                dmc.Tooltip(
                                                    dmc.Avatar(
                                                        ''.join([n[0] for n in reviewer.get('fullName').split()]),
                                                        radius = 'xl',
                                                        size = 'md',
                                                        color = f'rgb({np.random.randint(0,255)},{np.random.randint(0,255)},{np.random.randint(0,255)})'
                                                    ),
                                                    label = reviewer.get('fullName'),
                                                    position = 'bottom'
                                                ),
                                                href = f'https://orcid.org/{reviewer.get("orc_id")}',
                                                target = '_blank'
                                            )
                                            for reviewer in table_attribution['reviewers']
                                        ]
                                    )
                                )]
                            )
                        ],style = {'marginBottom':'5px'})
                    ]

                    organ_dash_table = [
                        dash_table.DataTable(
                            id = {'type':f'{self.component_prefix}-hra-viewer-table','index': 0},
                            columns = [{'name':i,'id':i,'deletable':False,'selectable':True} for i in organ_table.columns],
                            data = organ_table.to_dict('records'),
                            editable = False,
                            filter_action='native',
                            sort_action='native',
                            sort_mode='multi',
                            style_table = {
                                'overflowX': 'auto',
                                'maxWidth': '800px'
                            },
                            tooltip_data = [
                                {
                                    column: {'value': str(value),'type':'markdown'}
                                    for column,value in row.items()
                                } for row in organ_table.to_dict('records')
                            ],
                            tooltip_duration = None
                        )
                    ]
                    
                    data_doi = dbc.Row([
                            dbc.Col(
                                dmc.NavLink(
                                    label = html.Div(f'Data DOI: {table_attribution.get("doi","").split("/")[-1]}',style={'align':'center'}),
                                    href = table_attribution.get('doi',''),
                                    target = '_blank'
                                ),
                                md = 12
                            )
                        ],align = 'center')
                

                    viewer_children = dbc.Row([
                        dbc.Col([
                            html.Div([
                                html.Div(reviewers_and_authors),
                                html.Hr(),
                                dbc.Row([
                                    dbc.Col([
                                        html.Div(
                                            organ_dash_table,
                                            style = {'maxHeight': '500px','overflow': 'scroll','width': '100%'}
                                        )
                                    ],md = 'auto')
                                ]),
                                html.Hr(),
                                dbc.Row([
                                    dbc.Col([
                                        dbc.Label(
                                            html.H6('Table Data Sources:'),
                                            html_for={'type':f'{self.component_prefix}-hra-viewer-table-sources','index':0}
                                        )
                                    ],md=12)
                                ]),
                                data_doi,
                                dbc.Row([
                                    dbc.Col(
                                        dbc.Label('Date: ',html_for = {'type': f'{self.component_prefix}-hra-viewer-date','index': 0}),
                                        md = 4
                                    ),
                                    dbc.Col(
                                        table_attribution.get('creation_data',''),
                                        md = 8
                                    )
                                ],align='center'),
                                dbc.Row([
                                    dbc.Col(
                                        dbc.Label('Version Number: ',html_for = {'type':f'{self.component_prefix}-hra-viewer-version','index': 0}),
                                        md = 4
                                    ),
                                    dbc.Col(
                                        table_attribution.get('version',''),
                                        md = 8
                                    )
                                ],align='center')
                            ])
                        ],
                        width = True)
                    ],style={'width':'100%'})

                else:
                    viewer_children = dbc.Alert(f'Unable to get ASCT+B Table for {viewer_drop_value}',color='warning')
            else:
                viewer_children = dbc.Alert(f'Unable to get ASCT+B Table for {viewer_drop_value}',color='warning')

        return [viewer_children]

    def refresh_asctb_tables(self, refresh_clicked):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        try:
            self.asct_b_release = self.get_latest_asctb()

            self.organ_table_options = [
                {'label': f'{i} ASCT+B Table', 'value': i, 'disabled': False}
                for i in self.asct_b_release['Organ'].tolist()
            ]

            return_options = {'label': 'FTU Explorer','value': 'FTU Explorer','disabled': False} + self.organ_table_options
            return_error_div = dbc.Alert(
                'Success',
                color = 'success',
                dismissable=True
            )
        except Exception as e:
            print(e)
            return_options = no_update
            return_error_div = dbc.Alert(
                color = 'danger',
                children = [
                    dbc.Row([
                        dbc.Col([
                            f'Error loading ASCT+B Tables'
                        ],md = 8),
                        dbc.Col([
                            html.A(
                                html.I(
                                    className = 'fa-solid fa-rotate fa-2x',
                                    n_clicks = 0,
                                    id = {'type': f'{self.component_prefix}-hra-viewer-refresh-icon','index': 0}
                                )
                            ),
                            dbc.Tooltip(
                                target = {'type': f'{self.component_prefix}-hra-viewer-refresh-icon','index': 0},
                                children = 'Click to re-try grabbing ASCT+B tables'
                            )],
                            md = 4
                        )
                    ])
                ]
            )
    

        return [return_error_div],[return_options]











"""

Structure schemas for different items in SQLite database

"""
import json
import uuid
import time

from datetime import datetime

from sqlalchemy import (
    not_, func, select, create_engine, update,
    Column, String, Boolean,ForeignKey, JSON)
from sqlalchemy.orm import declarative_base, sessionmaker, mapped_column, Session, scoped_session
from sqlalchemy.pool import NullPool

from typing import Generator
from contextlib import contextmanager
from shapely.geometry import box, shape

from typing_extensions import Union

from .models import Base, User, VisSession, Item, Layer, Structure, ImageOverlay, Annotation


TABLE_NAMES = {
    'user': User,
    'visSession': VisSession,
    'item': Item,
    'layer': Layer,
    'structure': Structure,
    'image_overlay': ImageOverlay,
    'annotation': Annotation
}


class fusionDB:
    def __init__(self,
                 db_url:str,
                 echo:bool = False):
        
        self.engine = create_engine(
            db_url,
            connect_args={
                "check_same_thread": False
            },
            echo = echo,
            pool_pre_ping = True,
            pool_recycle=3600
        )

        Base.metadata.create_all(bind = self.engine)
        self.SessionLocal = scoped_session(sessionmaker(bind=self.engine))

    @contextmanager
    def get_db(self) -> Generator[Session, None, None]:
        # Generator type has types, yield, send, return (in this case it yields a Session, sends None and returns None)
        db: Session = self.SessionLocal()
        try:
            yield db
            db.commit()
        except Exception:
            db.rollback()
            raise
        finally:
            db.close()

    def get_uuid(self):
        return uuid.uuid4().hex[:24]
    
    def add(self, obj, session = None):
        # Using sessionmaker in a context manager 
        # (closes automatically and rolls back in the event of a database error)
        if session is None:
            with self.get_db() as session:
                session.add(obj)
                session.commit()
        else:
            session.add(obj)

        return True
    
    def get_create(self, table_name:str, inst_id:Union[str,None] = None, kwargs:Union[dict,None] = None):
        
        with self.get_db() as session:
            if table_name in TABLE_NAMES:
                if not inst_id is None:
                    get_create_result = session.query(
                        TABLE_NAMES.get(table_name)
                    ).filter_by(id = inst_id).first()

                    if not get_create_result:
                        #TODO: Add a "created" option to each model

                        # This is if this thing does not exist in the table
                        updated = datetime.now()
                        get_create_result = TABLE_NAMES.get(table_name)(
                            id = inst_id,
                            **kwargs | {'updated': updated}
                        )

                        self.add(get_create_result, session)
                    else:
                        # This is if this thing DOES exist in the table, update
                        updated = datetime.now()
                        get_create_result = TABLE_NAMES.get(table_name)(
                            id = inst_id,
                            **kwargs | {'updated': updated}
                        )
                        update_kwargs = kwargs | {'updated': updated}
                        session.execute(update(TABLE_NAMES.get(table_name)).where(getattr(TABLE_NAMES.get(table_name),'id')==inst_id).values(update_kwargs))
                
                else:
                    new_id = self.get_uuid()
                    updated = datetime.now()
                    get_create_result = TABLE_NAMES.get(table_name)(
                        id = new_id,
                        **kwargs | {'updated': updated}
                    )

                    self.add(get_create_result, session)

                session.commit()

                return get_create_result
            else:
                return None

    def count(self, table_name:str):
        """Get the count of unique instances within this specific table

        :param table_name: Name of table to query (user, visSession, item, layer, structure, image_overlay, annotation)
        :type table_name: str
        :return: Count of instances
        :rtype: int
        """

        #TODO: Add some filtering capability here
        with self.get_db() as session:
            if table_name in TABLE_NAMES:
                count_result = session.execute(
                    select(func.count(func.distinct(TABLE_NAMES.get(table_name).id)))
                ).one()

                return count_result[0]
            else:
                return 0

    def search_op(self, search_query, table_name:str, column:str, op:dict):
        """Method for applying different types of filters to database queries

        :param search_query: DB query object
        :type search_query: None
        :param table_name: Name of table being searched
        :type table_name: str
        :param column: Name of column being queried
        :type column: str
        :param op: Operation dictionary
        :type op: dict
        :return: Updated DB query object
        :rtype: None
        """
        op_type = list(op.keys())[0]
        op_query = list(op.values())
        if op_type in ['==','!=','>','<','>=','<=','in','!in']:
            # Numeric operations
            if all([type(i) in [int,float] for i in op_query]) and op_type in ['==','!=','>','<','>=','<=']:
                if op_type=='==':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column)==op_query[0]
                    )
                elif op_type=='!=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column)!=op_query[0]
                    )
                elif op_type=='>':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) > op_query[0]
                    )
                elif op_type=='>=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) >= op_query[0]
                    )
                elif op_type=='<':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) < op_query[0]
                    )
                elif op_type=='<=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) <= op_query[0]
                    )

            # String operations
            elif all([type(i)==str for i in op_query]) and op_type in ['==','!=']:
                if op_type=='==':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) == op_query[0]
                    )
                elif op_type=='!=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) != op_query[0]
                    )
            
            # List operations
            elif all([type(i)==list for i in op_query]) and op_type in ['in','!in']:
                if op_type=='in':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column).in_(op_query[0])
                    )
                elif op_type=='!in':
                    search_query = search_query.filter(
                        not_(getattr(TABLE_NAMES.get(table_name),column).in_(op_query[0]))
                    )

            return search_query

        else:
            return search_query

    async def search(self, search_kwargs:dict, size:Union[int,None] = None, offset = 0, order = None):
        """Search DB

        :param search_kwargs: Dictionary containing "type" (table name) and "filters"
        :type search_kwargs: dict
        :param size: Number of instances to return, if None returns all that match the search, defaults to None
        :type size: int, optional
        :param offset: Offset of instances to return from search, defaults to 0
        :type offset: int, optional
        :param order: Name of column to order by (Not Implemented), defaults to None
        :type order: None, optional
        :return: Results of DB search
        :rtype: list
        """

        if not search_kwargs.get('type','') in TABLE_NAMES:
            return []

        #print(f'db search called: {json.dumps(search_kwargs,indent=4)}')
        with self.get_db() as session:
            search_query = session.query(
                TABLE_NAMES.get(search_kwargs.get('type'))
            )
            
            # Applying filters
            search_filters = search_kwargs.get('filters')
            if not search_filters is None:
                
                table_joins = list(set(list(search_filters.keys())))
                for t in table_joins:
                    if t in TABLE_NAMES:
                        if not t == search_kwargs.get('type'):
                            search_query = search_query.join(TABLE_NAMES.get(t))

                        for k,v in search_filters.get(t).items():
                            if type(v)==str:
                                search_query = search_query.filter(
                                    getattr(TABLE_NAMES.get(t),k)==v
                                )
                            elif type(v)==dict:
                                search_query = self.search_op(search_query,t,k,v)

            return_list = []
            for idx,i in enumerate(search_query.all()):
                if not size is None:
                    if len(return_list)==size:
                        break

                if idx>=offset:
                    return_list.append(i.to_dict())

            #print(f'n returned from db_search: {len(return_list)}')
            
            return return_list
      
    def add_slide(self, slide_id:str, slide_name:str, metadata: dict, image_metadata: dict, image_filepath:Union[str,None], annotations_metadata:dict,annotations:Union[list,dict,None]):

        new_item = self.get_create(
            table_name = 'item',
            inst_id = slide_id,
            kwargs = {
                'name': slide_name,
                'meta': metadata,
                'image_meta': image_metadata,
                'ann_meta': annotations_metadata,
                'filepath': image_filepath
            }
        )

        for ann_idx, ann in enumerate(annotations):
            # Adding layer
            new_layer = self.get_create(
                table_name = 'layer',
                inst_id = ann['properties']['_id'],
                kwargs = {
                    'name': ann['properties']['name'],
                    'item': slide_id
                }
            )

            if 'image_path' in ann:
                # Adding image overlay layer
                new_image_overlay = self.get_create(
                    table_name = 'image_overlay',
                    inst_id = ann['properties']['_id'],
                    kwargs = {
                        'bounds': ann['image_bounds'],
                        'properties': ann['image_properties'],
                        'image_src': ann['image_path'],
                        'layer': ann['properties']['_id']
                    }
                )
            else:
                # Adding Structures in Layer
                for f_idx, f in enumerate(ann['features']):
                    new_structure = self.get_create(
                        table_name = 'structure',
                        inst_id = f['properties']['_id'],
                        kwargs = {
                            'geom': f['geometry'],
                            'properties': f['properties'],
                            'layer': ann['properties']['_id'],
                            'item': slide_id
                        }
                    )

    def add_vis_session(self, vis_session: dict):
        #TODO: Adding new visualization session to database (local should just be the same each time so don't have to back it up)
        # Contains {"current": [], "local":[], "data": {}, "current_user": {}, "session": {}}

        #TODO: Guest user sessions all start with "guestuser"
        vis_session_kwargs = {
            'user': vis_session.get('current_user',{'_id': f'guestuser'+self.get_uuid()[:15]}).get('_id'),
            'data': vis_session.get('data',{})
        }

        new_vis_session = self.get_create(
            table_name = 'visSession',
            inst_id = vis_session.get('session',{}).get('id'),
            kwargs = vis_session_kwargs
        )

        #TODO: Adding items in "current"
        for c in vis_session.get('current',[]):
            print(f'add_vis_session current slide: {json.dumps(c,indent=4)}')
            item_kwargs = {
                'name': c.get('name'),
                'meta': c.get('meta'),
                'image_meta': c.get('image_meta'),
                'ann_meta': c.get('ann_meta'),
                'filepath': c.get('filepath'),
                'session': vis_session.get('session',{}).get('id')
            }
            new_item = self.get_create(
                table_name = 'item',
                inst_id = c.get('id'),
                kwargs = c
            )

            #TODO: Get Layers, Structures, ImageOverlays, and Annotations? Or wait to add those to the database?
            # Should duplicates be added? It would be important to preserve different versions of each Layer/Structure/etc. if changes are made.
            # The default/static version of annotations can just be the source (local file/DSA annotations/item/{id})

    def get_names(self, table_name:str, size:Union[int,None]=None, offset = 0):

        return_names = []
        with self.get_db() as session:

            if table_name in TABLE_NAMES:
                name_query = session.execute(
                    select(getattr(TABLE_NAMES.get(table_name),'name'))
                )

                for a_idx,a in enumerate(name_query.all()):
                    if not size is None:
                        if len(return_names)==size:
                            break
                    if a_idx>=offset:
                        return_names.append(a[0])

            return return_names

    async def get_item_annotations(self, item_id:str, user_id:Union[str,None] = None, vis_session_id:Union[str,None] = None)->list:
        """Loading annotations from item database

        :param item_id: String uuid for an item
        :type item_id: str
        :param user_id: String uuid for a user, defaults to None
        :type user_id: Union[str,None], optional
        :param vis_session_id: String uuid for a visualization session, defaults to None
        :type vis_session_id: Union[str,None], optional
        :return: List of GeoJSON-formatted FeatureCollections (and image overlays if present)
        :rtype: list
        """

        item_annotations = []

        item_layers = await self.search(
            search_kwargs = {
                'type': 'layer',
                'filters': {
                    'item': {
                        'id': item_id
                    }
                }
            }
        )

        for l in item_layers:
            layer_name = l.get('name')
            layer_id = l.get('id')

            layer_structures = await self.search(
                search_kwargs={
                    'type': 'structure',
                    'filters': {
                        'layer': {
                            'id': layer_id
                        }
                    }
                }
            )

            if len(layer_structures)>0:
                item_annotations.append(
                    {
                        'type': 'FeatureCollection',
                        'properties': {
                            'name': layer_name,
                            '_id': layer_id
                        },
                        'features': [
                            {
                                'type': 'Feature',
                                'geometry': s.get('geom'),
                                'properties': s.get('properties')
                            }
                            for s in layer_structures
                        ]
                    }
                )
            else:
                # This could be an ImageOverlay layer
                image_overlays = await self.search(
                    search_kwargs = {
                        'type': 'image_overlay',
                        'filters': {
                            'layer': {
                                'id': layer_id
                            }
                        }
                    }
                )

                if len(image_overlays)>0:
                    item_annotations.extend(
                        [
                            i.to_dict() for i in image_overlays
                        ]
                    )

        return item_annotations

    def get_structure_property_keys(self, item_id:Union[str,list,None] = None, layer_id:Union[str,list,None] = None):
        #TODO: Make a more efficient way to get names of properties for each structure
        pass

    def get_structure_property_data(self, item_id:Union[str,list,None] = None, layer_id:Union[str,list,None] = None, structure_id:Union[str,list,None] = None, property_list:Union[str,list] = None):
        """Extracting one or multiple properties from structures given id filters.

        :param item_id: String uuid for one or multiple image items, defaults to None
        :type item_id: Union[str,list,None], optional
        :param layer_id: String uuid for one or multiple layers, defaults to None
        :type layer_id: Union[str,list,None], optional
        :param structure_id: String uuid for one or multiple structures, defaults to None
        :type structure_id: Union[str,list,None], optional
        :param property_list: List of one or more properties to extract from structures, defaults to None
        :type property_list: Union[str,list], optional
        :return: Records-formatted list of dictionaries with each property, along with structure id, bounding box (minx, miny, maxx, maxy), layer id, layer name, item id, and item name
        :rtype: list
        """

        if property_list is None:
            return []
        elif type(property_list)==str:
            property_list = [property_list]

        # Ensuring uniqueness of property names
        property_list = list(set(property_list))

        with self.get_db() as session:
            search_query = session.query(
                *[Structure.properties[p.split(' --> ')] for p in property_list],
                Structure.id,
                Structure.geom,
                Layer.id,
                Layer.name,
                Item.id,
                Item.name
            ).filter(Structure.layer == Layer.id).filter(Layer.item==Item.id)

            if not item_id is None:
                if type(item_id)==list:
                    search_query = search_query.filter(Item.id.in_(item_id))
                elif type(item_id)==str:
                    search_query = search_query.filter(Item.id == item_id)

            if not layer_id is None:
                if type(layer_id)==list:
                    search_query = search_query.filter(Layer.id.in_(layer_id))
                elif type(layer_id)==str:
                    search_query = search_query.filter(Layer.id==layer_id)

            if not structure_id is None:
                if type(structure_id)==list:
                    search_query = search_query.filter(Structure.id.in_(structure_id))
                elif type(structure_id)==str:
                    search_query = search_query.filter(Structure.id==structure_id)
            
            returned_props = property_list + ['structure.id','geometry','layer.id','layer.name','item.id','item.name']
            return_list = []
            for idx,i in enumerate(search_query.all()):
                i_dict = {'_index': idx}
                for prop,prop_name in zip(i,returned_props):
                    if not prop_name=='geometry':
                        i_dict[prop_name] = prop
                    else:
                        i_dict['bbox'] = list(shape(prop).bounds)

                return_list.append(i_dict)

            return return_list

    def get_structures_in_bbox(self, bbox:list, item_id:Union[str,None] = None, layer_id:Union[str,list,None] = None, structure_id:Union[str,list,None] = None):
        """Querying database for structures that intersect with a 

        :param bbox: _description_
        :type bbox: list
        :param item_id: _description_, defaults to None
        :type item_id: Union[str,None], optional
        :param layer_id: _description_, defaults to None
        :type layer_id: Union[str,list,None], optional
        :param structure_id: _description_, defaults to None
        :type structure_id: Union[str,list,None], optional
        :return: _description_
        :rtype: _type_
        """
        
        #TODO: Test the performance of this using shape().intersects() vs. checking min/max ranges for bounding boxes
        with self.get_db() as session:
            start = time.time()
            search_query = session.query(
                Structure.id,
                Structure.geom
            )

            if not item_id is None:
                if type(item_id)==str:
                    search_query = search_query.filter(Item.id == item_id)

            if not layer_id is None:
                if type(layer_id)==list:
                    search_query = search_query.filter(Layer.id.in_(layer_id))
                elif type(layer_id)==str:
                    search_query = search_query.filter(Layer.id==layer_id)

            if not structure_id is None:
                if type(structure_id)==list:
                    search_query = search_query.filter(Structure.id.in_(structure_id))
                elif type(structure_id)==str:
                    search_query = search_query.filter(Structure.id==structure_id)

            # Box should be minx, miny, maxx, maxy
            query_box = box(*bbox)
            return_list = []
            for idx,i in enumerate(search_query.all()):
                structure_id = i[0]
                structure_geom = i[1]
                if shape(structure_geom).intersects(query_box):
                    return_list.append(structure_id)


            print(f'Time for get_structures_in_bbox: {time.time() - start}')


            return return_list





"""
Defining layout for a "Welcome" page in FUSION containing embedded YouTube videos with documentation/how-tos
as well as a brief description of the tool and 

"""

import os

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import dcc, ALL, MATCH, ctx, exceptions, no_update
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash_extensions.enrich import (
    DashBlueprint,
    html, 
    MultiplexerTransform, 
    PrefixIdTransform, 
    Input, State, Output)

from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools import Tool

class WelcomePage(Tool):
    def __init__(self):
        super().__init__()

        #TODO: Just a note, these image addresses get rate-limited pretty quickly from GitHub, ideally these 
        # images would be locally accessible but I think that would make the package much heavier.
        self.content_dict = {
            'Background': html.Div([
                dbc.Row(html.H5('Background Information')),
                html.Hr(),
                dbc.Row(
                    'FUSION is an interactive analysis tool which enables integrated analysis of histology images with or without spatial --omics data.',
                    style={'marginBottom':'5px','marginTop':'5px'}
                ),
                dbc.Row([
                    dbc.Col(
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/background-flowchart.png?raw=true',
                            alt = 'FUSION background diagram',
                            style = {
                                'width': '100%'
                            }
                        ),
                        md = 6
                    ),
                    dbc.Col([
                            'This cloud implementation of FUSION demonstrates the ability of FUSION to leverage cloud resources for storage and running jobs with high computational complexity. ',
                            'Digital Slide Archive (DSA) is here used as a "backend" for organization of files and running Dockerized plugins formatted using Slicer CLI specifications. '
                        ],md = 6)
                    ], align='center'
                ),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/page-menu.png?raw=true',
                            alt = 'FUSION PageMenu',
                            style = {
                                'width': '100%'
                            }
                        )
                    ],md=6),
                    dbc.Col([
                        'Navigate to different pages in FUSION using the PAGES MENU in the navigation bar at the top of your screen. ',
                        'There you will see a few different buttons with page addresses on them like this: "/APP/VISUALIZATION", "/APP/DATASET BUILDER", etc. ',
                        'The Visualization and Dataset Builder pages are openly accessible to everyone. ',
                        'The Dataset Uploader page requires creating an account in order to associate any uploaded data to each user\'s quota. ',
                        'You can make a new account (100% free!), using the "DSA LOGIN" button underneath the navigation bar. ',
                        'Follow the instructions to create a new account or login using your username and password and navigate to the Dataset Uploader page to start uploading your own data. '
                    ],md=6)
                ])
            ]),
            'The SlideMap Component': html.Div([
                dbc.Row(html.H5('The SlideMap Component')),
                html.Hr(),
                dbc.Row(
                    'The SlideMap component is a base class which contains interactive components for high-resolution images. It leverages Dash-Leaflet for generation of Map-style interactive components integrating with Dash callbacks with other components.'
                ),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/slidemap/slidemap-tile-coordinates.png?raw=true',
                            alt = 'High-resolution tiled images',
                            style = {
                                'width': '100%'
                            }
                        ),
                    ], md = 6),
                    dbc.Col([
                        'Since high-resolution images are very large files, they are typically not loaded into memory at the highest resolution all at once, especially for web-applications. ',
                        'Therefore, smaller tiles are accessed from varying resolution levels using an application programming interface (API) request. ',
                        'For cloud-stored images in FUSION, this is facilitated by the large-image Girder plugin. ',
                    ],md = 6)
                ],style={'marginBottom':'5px','marginTop':'5px'}),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/slidemap/slidemap-dropdown.png?raw=true',
                            alt = 'SlideMap dropdown',
                            style = {
                                'width': '100%'
                            }
                        )
                    ], md = 6),
                    dbc.Col([
                        'To load a slide in FUSION, first navigate to the visualization page containing the SlideMap component. ',
                        'Next, select the slide you would like to view from the dropdown menu at the top of the SlideMap component. ',
                        'This will trigger FUSION collecting the necessary rendering information for that slide, including the tiles URL, annotations, and any other slide metadata. '
                    ])
                ]),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/slidemap/slidemap-layers.png?raw=true',
                            alt = 'SlideMap layers',
                            style = {
                                'width': '100%'
                            }
                        )
                    ],md=6),
                    dbc.Col([
                        'Use the layers selector in the top right of the map to turn different overlaid annotations off and on. '
                    ],md=6)
                ],align='center',style = {'marginBottom':'5px'}),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/slidemap/slidemap-multiframe.png?raw=true',
                            alt = 'MxIF image channels',
                            style = {
                                'width': '100%'
                            }
                        )
                    ],md=6),
                    dbc.Col([
                        'Multi-plexed immunofluorescence images which contain multiple channels for each fluorescent marker, can either be viewed individually in grayscale, or combined with manually selected colors using the ChannelMixer component. '
                    ],md=6)
                ],align='center',style = {'marginBottom':'5px'})
            ]),
            'Structure-level Properties': html.Div([
                dbc.Row(html.H5('Structure-level Properties')),
                html.Hr(),
                dbc.Row(
                    'One of the key concepts behind FUSION is the incorporation of structure-level properties. These comprise any kind of derived data or identifying information for structures in a slide which can be used in downstream analysis or visualization',
                    style={'marginBottom':'5px','marginTop':'5px'}
                ),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/geojson-example.png?raw=true',
                            alt = 'GeoJSON example',
                            style = {
                                'width':'100%'
                            }
                        )
                    ],md=6),
                    dbc.Col([
                        'Following GeoJSON conventions, each annotation is defined as a FeatureCollection consisting of one or more Features containing a type, geometry, and properties field. ',
                        'FUSION accesses structure-level features from the properties field and incorporates them into visualizations and analyses dynamically. '
                    ],md=6)
                ],align='center')
            ]),
            'Accessing Data with Tools': html.Div([
                dbc.Row(html.H5('Accessing Data with Tools')),
                html.Hr(),
                dbc.Row(
                    'Data for each slide can be accessed through a variety of interactive components in the "Tools" tabs to the right of the SlideMap.',
                    style={'marginBottom':'5px','marginTop':'5px'}
                ),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/properties/property-accessors.png?raw=true',
                            alt = 'Property accessors in FUSION',
                            style = {
                                'width':'100%'
                            }
                        )
                    ],md=6),
                    dbc.Col([
                        'The GlobalPropertyPlotter, PropertyPlotter, and PropertyViewer components access structural properties at 3 different levels. ',
                        'The PropertyViewer component loads properties from only the structures within the current viewport and can be used for assessing whether a given property is localized to particular regions. ',
                        'The PropertyPlotter accesses all of the properties for all of the structures in the current slide regardless of where they are. ',
                        'The GlobalPropertyPlotter accesses all of the properties from all of the structures from all of the slides in your current Visualization Session. To modify the slides in your Visualization Session, use the Dataset Builder page. '
                    ],md=6)
                ])
            ]),
            'Getting Started with fusion-tools': html.Div([
                dbc.Row(html.H5('Getting Started with fusion-tools')),
                html.Hr(),
                dbc.Row(
                    'This FUSION visualization is constructed using the fusion-tools Python library, which includes a variety of additional functionality for generating customizable dashboards locally.',
                    style={'marginBottom':'5px','marginTop':'5px'}
                ),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/fusion-tools/fusion-layout.png?raw=true',
                            alt = 'FUSION layout in fusion-tools',
                            style = {
                                'width':'100%'
                            }
                        )
                    ],md=6),
                    dbc.Col([
                        'The Visualization object in fusion-tools lets users define which data they would like to initially load as well as which components should be added to the layout. ',
                        'Multi-page layouts can be created by passing a dictionary with the name of each page to the "components" argument. ',
                        'The "start()" method initializes the application server where it is then accessible at the localhost:\{port\} address. '
                    ],md=6)
                ]),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        html.Img(
                            src = 'https://github.com/spborder/fusion-tools/blob/main/docs/images/fusion-tools/simple-layout.png?raw=true',
                            alt= 'simple fusion-tools layout',
                            style = {
                                'width': '100%'
                            }
                        )
                    ],md=8),
                    dbc.Col([
                        'Simpler layouts which include locally stored data can be defined using a subset of available components. ',
                        'You can also create FUSION visualizations inline with Jupyter Notebooks by adding "jupyter": True to the "app_options" argument. '
                    ],md=4)
                ])
            ])
        }

    def __str__(self):
        return "Welcome Page"

    def load(self,component_prefix:int):

        self.component_prefix = component_prefix
        
        self.title = 'Welcome Page'
        self.blueprint = DashBlueprint(
            transforms = [
                PrefixIdTransform(prefix = f'{self.component_prefix}'),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

    def get_callbacks(self):
        
        self.blueprint.callback(
            [
                Input({'type': 'welcome-nav','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'welcome-page-content','index': ALL},'children')
            ]
        )(self.update_page_content)

    def gen_layout(self, session_data:dict):

        # the style arguments for the sidebar. We use position:fixed and a fixed width
        SIDEBAR_STYLE = {
            "position": "relative",
            "top": 0,
            "left": 0,
            "bottom": 0,
            'height': '100%',
            "width": "25rem",
            "padding": "2rem 1rem",
            "background-color": "#f8f9fa"
        }

        # the styles for the main content position it to the right of the sidebar and
        # add some padding.
        CONTENT_STYLE = {
            "margin-left": "16rem",
            "margin-right": "2rem",
            "padding": "2rem 1rem",
        }

        sidebar = html.Div(
            [
                #html.H4("Categories", className="display-4"),
                #html.Hr(),
                html.P("Select a topic to view more information", className="lead"),
                html.Hr(),
                dbc.Nav(
                    [
                        dbc.NavLink(
                            n,
                            id = {'type': 'welcome-nav','index': n_idx},
                            active='exact'
                        )
                        for n_idx,n in enumerate(self.content_dict)
                    ],
                    vertical=True,
                    pills=True,
                ),
            ],
            style=SIDEBAR_STYLE,
        )

        content = html.Div(id={'type': "welcome-page-content",'index': 0}, style=CONTENT_STYLE)

        layout =  html.Div([
            dbc.Card([
                dbc.Row([
                    dbc.Col(
                        sidebar,
                        md = 2
                    ),
                    dbc.Col(
                        content,
                        md = 10
                    )
                ])
            ])
        ])

        self.blueprint.layout = layout

    def update_page_content(self, clicked):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        nav_clicked = list(self.content_dict.keys())[ctx.triggered_id['index']]

        return [self.content_dict[nav_clicked]]
        














